import atexit
import datetime
import faulthandler
import os
import platform
import subprocess
import tempfile
import unittest

import netCDF4
import numpy as np

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm
from cfdm.read_write.exceptions import DatasetTypeError

warnings = False

# Set up temporary files
n_tmpfiles = 9
tmpfiles = [
    tempfile.mkstemp("_test_read_write.nc", dir=os.getcwd())[1]
    for i in range(n_tmpfiles)
]
[
    tmpfile,
    tmpfileh,
    tmpfileh2,
    tmpfileh3,
    tmpfilec,
    tmpfilec2,
    tmpfilec3,
    tmpfile0,
    tmpfile1,
] = tmpfiles


def _remove_tmpfiles():
    """Remove temporary files created during tests."""
    for f in tmpfiles:
        try:
            os.remove(f)
        except OSError:
            pass


atexit.register(_remove_tmpfiles)


class read_writeTest(unittest.TestCase):
    """Test the reading and writing of field constructs from/to disk."""

    f0 = cfdm.example_field(0)
    f1 = cfdm.example_field(1)

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.LOG_LEVEL("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows: cfdm.LOG_LEVEL('DEBUG')
        #
        # < ... test code ... >
        # cfdm.log_level('DISABLE')
        self.filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
        )

        self.string_filename = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "string_char.nc"
        )

        self.netcdf3_fmts = [
            "NETCDF3_CLASSIC",
            "NETCDF3_64BIT",
            "NETCDF3_64BIT_OFFSET",
            "NETCDF3_64BIT_DATA",
        ]
        self.netcdf4_fmts = ["NETCDF4", "NETCDF4_CLASSIC"]
        self.netcdf_fmts = self.netcdf3_fmts + self.netcdf4_fmts

    def test_write_filename(self):
        """Test the writing of a named netCDF file."""
        f = self.f0
        a = f.data.array

        cfdm.write(f, tmpfile)
        g = cfdm.read(tmpfile)

        with self.assertRaises(Exception):
            cfdm.write(g, tmpfile)

        self.assertTrue((a == g[0].data.array).all())

    def test_read_extra(self):
        """Test the cfdm.read 'extra' keyword."""
        # Test field keyword of cfdm.read
        filename = self.filename

        f = cfdm.read(filename)
        self.assertEqual(len(f), 1)

        f = cfdm.read(
            filename, extra=["dimension_coordinate"], warnings=warnings
        )
        self.assertEqual(len(f), 4)

        f = cfdm.read(
            filename, extra=["auxiliary_coordinate"], warnings=warnings
        )
        self.assertEqual(len(f), 4)

        f = cfdm.read(filename, extra="cell_measure")
        self.assertEqual(len(f), 2)

        f = cfdm.read(filename, extra=["field_ancillary"])
        self.assertEqual(len(f), 4)

        f = cfdm.read(filename, extra="domain_ancillary", warnings=warnings)
        self.assertEqual(len(f), 4)

        f = cfdm.read(
            filename,
            extra=["field_ancillary", "auxiliary_coordinate"],
            warnings=warnings,
        )
        self.assertEqual(len(f), 7)

        self.assertEqual(
            len(
                cfdm.read(
                    filename,
                    extra=["domain_ancillary", "auxiliary_coordinate"],
                    warnings=warnings,
                )
            ),
            7,
        )
        self.assertEqual(
            len(
                cfdm.read(
                    filename,
                    extra=[
                        "domain_ancillary",
                        "cell_measure",
                        "auxiliary_coordinate",
                    ],
                    warnings=warnings,
                )
            ),
            8,
        )

        f = cfdm.read(
            filename,
            extra=(
                "field_ancillary",
                "dimension_coordinate",
                "cell_measure",
                "auxiliary_coordinate",
                "domain_ancillary",
            ),
            warnings=warnings,
        )
        self.assertEqual(len(f), 14)

    def test_read_write_format(self):
        """Test the cfdm.write 'fmt' keyword."""
        f = cfdm.read(self.filename)[0]
        for fmt in self.netcdf_fmts:
            cfdm.write(f, tmpfile, fmt=fmt)
            g = cfdm.read(tmpfile)
            self.assertEqual(len(g), 1)
            g = g[0]
            self.assertTrue(
                f.equals(g, verbose=3), f"Bad read/write of format: {fmt}"
            )

    def test_write_netcdf_mode(self):
        """Test the `mode` parameter to `write`, notably append mode."""
        g = cfdm.read(self.filename)  # note 'g' has one field

        # Test special case #1: attempt to append fields with groups
        # (other than 'root') which should be forbidden. Using fmt="NETCDF4"
        # since it is the only format where groups are allowed.
        #
        # Note: this is not the most natural test to do first, but putting
        # it before the rest reduces spurious seg faults for me, so...
        g[0].nc_set_variable_groups(["forecast", "model"])
        cfdm.write(g, tmpfile, fmt="NETCDF4", mode="w")  # 1. overwrite to wipe
        f = cfdm.read(tmpfile)
        with self.assertRaises(ValueError):
            cfdm.write(g[0], tmpfile, fmt="NETCDF4", mode="a")

        # Test special case #2: attempt to append fields with contradictory
        # featureType to the original file:
        g[0].nc_clear_variable_groups()
        g[0].nc_set_global_attribute("featureType", "profile")
        cfdm.write(
            g,
            tmpfile,
            fmt="NETCDF4",
            mode="w",
            global_attributes=("featureType", "profile"),
        )  # 1. overwrite to wipe
        h = cfdm.example_field(3)
        h.nc_set_global_attribute("featureType", "timeSeries")
        with self.assertRaises(ValueError):
            cfdm.write(h, tmpfile, fmt="NETCDF4", mode="a")
        # Now remove featureType attribute for subsquent tests:
        g_attrs = g[0].nc_clear_global_attributes()
        del g_attrs["featureType"]
        g[0].nc_set_global_attributes(g_attrs)

        # Set a non-trivial (i.e. not only 'Conventions') global attribute to
        # make the global attribute testing more robust:
        add_global_attr = ["remark", "A global comment."]
        original_global_attrs = g[0].nc_global_attributes()
        original_global_attrs[add_global_attr[0]] = None  # -> None on fields
        g[0].nc_set_global_attribute(*add_global_attr)

        # First test a bad mode value:
        with self.assertRaises(ValueError):
            cfdm.write(g[0], tmpfile, mode="g")

        for fmt in self.netcdf_fmts:  # test over all netCDF 3 and 4 formats
            # Other tests cover write as default mode (i.e. test with no mode
            # argument); here test explicit provision of 'w' as argument:
            cfdm.write(
                g,
                tmpfile,
                fmt=fmt,
                mode="w",
                global_attributes=add_global_attr,
            )
            f = cfdm.read(tmpfile)

            new_length = 1  # since 1 == len(g)
            self.assertEqual(len(f), new_length)
            # Ignore as 'remark' should be 'None' on the field as tested below
            self.assertTrue(f[0].equals(g[0], ignore_properties=["remark"]))
            self.assertEqual(
                f[0].nc_global_attributes(), original_global_attrs
            )

            # Main aspect of this test: testing the append mode ('a'): now
            # append all other example fields, to check a diverse variety.
            for ex_field_n, ex_field in enumerate(cfdm.example_fields()):
                # Note: after Issue #141, this skip can be removed.
                if ex_field_n == 1:
                    continue

                # Skip since "RuntimeError: Can't create variable in
                # NETCDF4_CLASSIC file from (2)  (NetCDF: Attempting netcdf-4
                # operation on strict nc3 netcdf-4 file)" i.e. not possible.
                if fmt == "NETCDF4_CLASSIC" and ex_field_n in (6, 7):
                    continue

                print(
                    "TODOUGRID: excluding example fields 8, 9, 10 until writing UGRID is enabled"
                )
                if ex_field_n in (8, 9, 10):
                    continue

                cfdm.write(ex_field, tmpfile, fmt=fmt, mode="a")
                f = cfdm.read(tmpfile)

                new_length += 1  # there should be exactly one more field now
                self.assertEqual(len(f), new_length)
                # Can't guarantee order of fields read in after the appends, so
                # check that the field is *somewhere* in the read-in fieldlist
                self.assertTrue(
                    any(
                        [
                            ex_field.equals(
                                file_field,
                                ignore_properties=[
                                    "comment",
                                    "featureType",
                                    "remark",
                                ],
                            )
                            for file_field in f
                        ]
                    )
                )
                for file_field in f:
                    self.assertEqual(
                        file_field.nc_global_attributes(),
                        original_global_attrs,
                    )

            # Now do the same test, but appending all of the example fields in
            # one operation rather than one at a time, to check that it works.
            cfdm.write(g, tmpfile, fmt=fmt, mode="w")  # 1. overwrite to wipe
            print(
                "TODOUGRID: excluding example fields 8, 9, 10 until writing UGRID is enabled"
            )
            append_ex_fields = cfdm.example_fields(*range(8))
            del append_ex_fields[1]  # note: can remove after Issue #141 closed
            # Note: can remove this del when Issue #140 is closed:
            if fmt in self.netcdf3_fmts:
                del append_ex_fields[5]  # n=6 ex_field, minus 1 for above del

            overall_length = len(append_ex_fields) + 1  # 1 for original 'g'
            cfdm.write(
                append_ex_fields, tmpfile, fmt=fmt, mode="a"
            )  # 2. now append
            f = cfdm.read(tmpfile)
            self.assertEqual(len(f), overall_length)

            # Also test the mode="r+" alias for mode="a".
            cfdm.write(g, tmpfile, fmt=fmt, mode="w")  # 1. overwrite to wipe
            cfdm.write(
                append_ex_fields, tmpfile, fmt=fmt, mode="r+"
            )  # 2. now append
            f = cfdm.read(tmpfile)
            self.assertEqual(len(f), overall_length)

            # The appended fields themselves are now known to be correct,
            # but we also need to check that any coordinates that are
            # equal across different fields have been shared in the
            # source netCDF, rather than written in separately.
            #
            # Note that the coordinates that are shared across the set of
            # all example fields plus the field 'g' from the contents of
            # the original file (self.filename) are as follows:
            #
            # 1. Example fields n=0 and n=1 share:
            #    <DimensionCoordinate: time(1) days since 2018-12-01 >
            # 2. Example fields n=0, n=2 and n=5 share:
            #    <DimensionCoordinate: latitude(5) degrees_north> and
            #    <DimensionCoordinate: longitude(8) degrees_east>
            # 3. Example fields n=2 and n=5 share:
            #    <DimensionCoordinate: air_pressure(1) hPa>
            # 4. The original file field ('g') and example field n=1 share:
            #    <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
            #    <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
            #    <Dimension...: atmosphere_hybrid_height_coordinate(1) >,
            #    <DimensionCoordinate: grid_latitude(10) degrees>,
            #    <DimensionCoordinate: grid_longitude(9) degrees> and
            #    <DimensionCoordinate: time(1) days since 2018-12-01 >
            #
            # Therefore we check all of those coordinates for singularity,
            # i.e. the same underlying netCDF variables, in turn.

            # But first, since the order of the fields appended isn't
            # guaranteed, we must find the mapping of the example fields to
            # their position in the read-in FieldList.
            f = cfdm.read(tmpfile)
            # Element at index N gives position of example field n=N in file
            file_field_order = []
            for ex_field in cfdm.example_fields():
                position = [
                    f.index(file_field)
                    for file_field in f
                    if ex_field.equals(
                        file_field,
                        ignore_properties=["comment", "featureType", "remark"],
                    )
                ]
                if not position:
                    position = [None]  # to record skipped example fields
                file_field_order.append(position[0])

            equal_coors = {
                ((0, "dimensioncoordinate2"), (1, "dimensioncoordinate3")),
                ((0, "dimensioncoordinate0"), (2, "dimensioncoordinate1")),
                ((0, "dimensioncoordinate1"), (2, "dimensioncoordinate2")),
                ((0, "dimensioncoordinate0"), (5, "dimensioncoordinate1")),
                ((0, "dimensioncoordinate1"), (5, "dimensioncoordinate2")),
                ((2, "dimensioncoordinate3"), (5, "dimensioncoordinate3")),
            }
            for coor_1, coor_2 in equal_coors:
                ex_field_1_position, c_1 = coor_1
                ex_field_2_position, c_2 = coor_2
                # Now map the appropriate example field to the file FieldList
                f_1 = file_field_order[ex_field_1_position]
                f_2 = file_field_order[ex_field_2_position]
                # None for fields skipped in test, distinguish from falsy 0
                if f_1 is None or f_2 is None:
                    continue
                self.assertEqual(
                    f[f_1]
                    .constructs()
                    .filter_by_identity(c_1)
                    .value()
                    .nc_get_variable(),
                    f[f_2]
                    .constructs()
                    .filter_by_identity(c_2)
                    .value()
                    .nc_get_variable(),
                )

            # Note: after Issue #141, the block below should be un-commented.
            #
            # The original file field 'g' must be at the remaining position:
            # rem_position = list(set(
            #     range(len(f))).difference(set(file_field_order)))[0]
            # # In the final cases, it is easier to remove the one differing
            # # coordinate to get the equal coordinates that should be shared:
            # original_field_coors = dict(f[rem_position].coordinates())
            # ex_field_1_coors = dict(f[file_field_order[1]].coordinates())
            # for orig_coor, ex_1_coor in zip(
            #         original_field_coors.values(), ex_field_1_coors.values()):
            #     # The 'auxiliarycoordinate2' construct differs for both, so
            #     # skip that but otherwise the two fields have the same coors:
            #     if orig_coor.identity == "auxiliarycoordinate2":
            #         continue
            #     self.assertEqual(
            #         orig_coor.nc_get_variable(),
            #         ex_1_coor.nc_get_variable(),
            #     )

            # Check behaviour when append identical fields, as an edge case
            #   1. Set up the fields and file to use to conduct this test
            g_new = cfdm.read(self.filename)[0]  # note 'g' has one field
            # There is an unresolved netcdf4-python issue when reading
            # VLEN arrays (see github.com/Unidata/netcdf4-python/issues/261)
            # so, until fixed, for NETCDF4 only we must delete the VLEN array
            # to conduct the test without it (but on a "kitchen sink" field).
            if fmt == "NETCDF4":
                aux = g_new.constructs.filter_by_property(
                    long_name="greek_letters"
                )
                g_new.del_construct(aux.key())
            g_copy = g_new.copy()
            cfdm.write(g_new, tmpfile, fmt=fmt, mode="w")  # overwrite to wipe

            #   2. Conduct the test by appending the identical field g_copy
            cfdm.write(g_copy, tmpfile, fmt=fmt, mode="a")
            f = cfdm.read(tmpfile)
            self.assertEqual(len(f), 2)  # i.e. len(f) == 2*len(g_new) == 2*1
            self.assertTrue(
                any(
                    [
                        file_field.equals(g_new, ignore_properties=["remark"])
                        for file_field in f
                    ]
                )
            )
            self.assertEqual(
                f[0].nc_global_attributes(), g_new.nc_global_attributes()
            )

    def test_read_write_compress_shuffle(self):
        """Test the `compress` and `shuffle` parameters to `write`."""
        f = self.f0.copy()
        f.data.nc_set_hdf5_chunksizes("contiguous")
        y = f.domain_axis("latitude")
        y.nc_set_unlimited(True)

        for fmt in ("NETCDF3_CLASSIC", "NETCDF4"):
            for shuffle in (
                False,
                True,
            ):
                for compress in (0, 1):
                    cfdm.write(
                        f, tmpfile, fmt=fmt, compress=compress, shuffle=shuffle
                    )
                    g = cfdm.read(tmpfile)[0]
                    self.assertTrue(f.equals(g))

            y.nc_set_unlimited(False)

    def test_read_write_missing_data(self):
        """Test reading and writing of netCDF with missing data."""
        f = cfdm.read(self.filename)[0]
        for fmt in self.netcdf_fmts:
            cfdm.write(f, tmpfile, fmt=fmt)
            g = cfdm.read(tmpfile)[0]
            self.assertTrue(
                f.equals(g, verbose=3), f"Bad read/write of format: {fmt}"
            )

    def test_read_mask(self):
        """Test reading and writing of netCDF with masked data."""
        f = self.f0.copy()

        N = f.size

        f.data[1, 1] = cfdm.masked
        f.data[2, 2] = cfdm.masked

        f.del_property("_FillValue", None)
        f.del_property("missing_value", None)

        cfdm.write(f, tmpfile)

        g = cfdm.read(tmpfile)[0]
        self.assertEqual(np.ma.count(g.data.array), N - 2)

        g = cfdm.read(tmpfile, mask=False)[0]
        self.assertEqual(np.ma.count(g.data.array), N)

        g.apply_masking(inplace=True)
        self.assertEqual(np.ma.count(g.data.array), N - 2)

        f.set_property("_FillValue", 999)
        f.set_property("missing_value", -111)
        cfdm.write(f, tmpfile)

        g = cfdm.read(tmpfile)[0]
        self.assertEqual(np.ma.count(g.data.array), N - 2)

        g = cfdm.read(tmpfile, mask=False)[0]
        self.assertEqual(np.ma.count(g.data.array), N)

        g.apply_masking(inplace=True)
        self.assertEqual(np.ma.count(g.data.array), N - 2)

    def test_write_datatype(self):
        """Test the `datatype` keyword argument to `write`."""
        f = cfdm.read(self.filename)[0]
        self.assertEqual(f.data.dtype, np.dtype(float))

        f.set_property("_FillValue", np.float64(-999.0))
        f.set_property("missing_value", np.float64(-999.0))

        cfdm.write(
            f,
            tmpfile,
            fmt="NETCDF4",
            datatype={np.dtype(float): np.dtype("float32")},
        )
        g = cfdm.read(tmpfile)[0]
        self.assertEqual(g.data.dtype, np.dtype("float32"))

        # Keyword single
        f = cfdm.read(self.filename)[0]
        self.assertEqual(f.dtype, np.dtype(float))
        cfdm.write(f, tmpfile, fmt="NETCDF4", single=True)
        g = cfdm.read(tmpfile)[0]
        self.assertEqual(g.dtype, np.dtype("float32"))

        # Keyword double
        f = g
        self.assertEqual(f.dtype, np.dtype("float32"))
        cfdm.write(f, tmpfile1, fmt="NETCDF4", double=True)
        g = cfdm.read(tmpfile1)[0]
        self.assertEqual(g.dtype, np.dtype(float))

        with self.assertRaises(Exception):
            cfdm.write(g, double=True, single=True)

        datatype = {np.dtype(float): np.dtype("float32")}
        with self.assertRaises(Exception):
            cfdm.write(g, datatype=datatype, single=True)

        with self.assertRaises(Exception):
            cfdm.write(g, datatype=datatype, double=True)

    def test_read_write_unlimited(self):
        """Test reading and writing with an unlimited dimension."""
        for fmt in self.netcdf_fmts:
            f = cfdm.read(self.filename)[0]
            domain_axes = f.domain_axes()

            domain_axes["domainaxis0"].nc_set_unlimited(True)
            cfdm.write(f, tmpfile, fmt=fmt, cfa=None)

            f = cfdm.read(tmpfile)[0]
            domain_axes = f.domain_axes()
            self.assertTrue(domain_axes["domainaxis0"].nc_is_unlimited())

        f = cfdm.read(self.filename)[0]

        domain_axes = f.domain_axes()

        domain_axes["domainaxis0"].nc_set_unlimited(True)
        domain_axes["domainaxis2"].nc_set_unlimited(True)
        cfdm.write(f, tmpfile, fmt="NETCDF4", cfa=None)

        f = cfdm.read(tmpfile)[0]
        domain_axes = f.domain_axes()
        self.assertTrue(domain_axes["domainaxis0"].nc_is_unlimited())
        self.assertTrue(domain_axes["domainaxis2"].nc_is_unlimited())

    def test_read_CDL(self):
        """Test the reading of files in CDL format."""
        subprocess.run(
            " ".join(["ncdump", self.filename, ">", tmpfile]),
            shell=True,
            check=True,
        )

        # For the cases of '-h' and '-c', i.e. only header info or coordinates,
        # notably no data, take two cases each: one where there is sufficient
        # info from the metadata to map to fields, and one where there isn't:
        #     1. Sufficient metadata, so should be read-in successfully
        subprocess.run(
            " ".join(["ncdump", "-h", self.filename, ">", tmpfileh]),
            shell=True,
            check=True,
        )
        subprocess.run(
            " ".join(["ncdump", "-c", self.filename, ">", tmpfilec]),
            shell=True,
            check=True,
        )

        #     2. Insufficient metadata, so should error with a message as such
        geometry_1_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "geometry_1.nc"
        )
        subprocess.run(
            " ".join(["ncdump", "-h", geometry_1_file, ">", tmpfileh2]),
            shell=True,
            check=True,
        )
        subprocess.run(
            " ".join(["ncdump", "-c", geometry_1_file, ">", tmpfilec2]),
            shell=True,
            check=True,
        )

        f0 = cfdm.read(self.filename)[0]

        # Case (1) as above, so read in and check the fields are as should be
        f = cfdm.read(tmpfile)[0]
        cfdm.read(tmpfileh)[0]
        c = cfdm.read(tmpfilec)[0]

        # Case (2) as above, so the right error should be raised on read
        with self.assertRaises(ValueError):
            cfdm.read(tmpfileh2)[0]

        with self.assertRaises(ValueError):
            cfdm.read(tmpfilec2)[0]

        self.assertTrue(f0.equals(f, verbose=3))

        self.assertTrue(
            f.construct("grid_latitude").equals(
                c.construct("grid_latitude"), verbose=3
            )
        )
        self.assertTrue(
            f0.construct("grid_latitude").equals(
                c.construct("grid_latitude"), verbose=3
            )
        )

        with self.assertRaises(DatasetTypeError):
            cfdm.read("test_read_write.py")

        # TODO: make portable instead of skipping on Mac OS (see Issue #25):
        #       '-i' aspect solved, but the regex patterns need amending too.
        if platform.system() != "Darwin":  # False for Mac OS(X) only
            for regex in [
                r'"1 i\ \ "',
                r'"1 i\// comment"',
                r'"1 i\ // comment"',
                r'"1 i\ \t// comment"',
            ]:
                # Note that really we just want to do an in-place sed
                # ('sed -i') but because of subtle differences between the
                # GNU (Linux OS) and BSD (some Mac OS) command variants a
                # safe portable one-liner may not be possible. This will
                # do, overwriting the intermediate file. The '-E' to mark
                # as an extended regex is also for portability.
                subprocess.run(
                    " ".join(
                        [
                            "sed",
                            "-E",
                            "-e",
                            regex,
                            tmpfileh,
                            ">" + tmpfileh3,
                            "&&",
                            "mv",
                            tmpfileh3,
                            tmpfileh,
                        ]
                    ),
                    shell=True,
                    check=True,
                )

                cfdm.read(tmpfileh)[0]

        # Finally test an invalid CDL input
        with open(tmpfilec3, "w") as fh:
            fh.write("netcdf test_file {\n  add badness\n}")
        # TODO: work out (if it is even possible in a farily simple way) how
        # to suppress the expected error in stderr of the ncdump command
        # called by cfdm.read under the hood. Note that it can be easily
        # suppressed at subprocess call-time (but we don't want to do that in
        # case of genuine errors) and the following doesn't work as it doesn't
        # influence the subprocess: with contextlib.redirect_stdout(os.devnull)
        with self.assertRaises(RuntimeError):
            cfdm.read(tmpfilec3)

    def test_read_write_string(self):
        """Test the `string` keyword argument to `read` and `write`."""
        fN = cfdm.read(self.string_filename, netcdf_backend="netCDF4")
        fH = cfdm.read(self.string_filename, netcdf_backend="h5netcdf")

        n = int(len(fN) / 2)

        for i in range(0, n):
            j = i + n
            self.assertTrue(fN[i].data.equals(fN[j].data, verbose=3))
            self.assertTrue(fN[j].data.equals(fN[i].data, verbose=3))

        # Check that netCDF4 and h5netcdf give the same results
        for i, j in zip(fN, fH):
            self.assertTrue(i.data.equals(j.data))

        # Note: Don't loop round all netCDF formats for better
        #       performance. Just one netCDF3 and one netCDF4 format
        #       is sufficient to test the functionality

        f0 = cfdm.read(self.string_filename)
        for string0 in (True, False):
            for fmt0 in ("NETCDF4", "NETCDF3_CLASSIC"):
                cfdm.write(f0, tmpfile0, fmt=fmt0, string=string0)

                for string1 in (True, False):
                    for fmt1 in ("NETCDF4", "NETCDF3_CLASSIC"):
                        cfdm.write(f0, tmpfile1, fmt=fmt1, string=string1)

                        for i, j in zip(
                            cfdm.read(tmpfile1), cfdm.read(tmpfile0)
                        ):
                            self.assertTrue(i.equals(j, verbose=3))

    def test_read_write_Conventions(self):
        """Test the `Conventions` keyword argument to `write`."""
        f = cfdm.read(self.filename)[0]

        version = "CF-" + cfdm.CF()
        other = "ACDD-1.3"

        for Conventions in (other,):
            cfdm.write(f, tmpfile0, Conventions=Conventions)
            g = cfdm.read(tmpfile0)[0]
            self.assertEqual(
                g.get_property("Conventions"),
                " ".join([version, other]),
                f"{g.get_property('Conventions')!r}, {Conventions!r}",
            )

        for Conventions in (version, "", " ", ",", ", "):
            Conventions = version
            cfdm.write(f, tmpfile0, Conventions=Conventions)
            g = cfdm.read(tmpfile0)[0]
            self.assertEqual(
                g.get_property("Conventions"),
                version,
                f"{g.get_property('Conventions')!r}, {Conventions!r}",
            )

        for Conventions in ([version], [version, other]):
            cfdm.write(f, tmpfile0, Conventions=Conventions)
            g = cfdm.read(tmpfile0)[0]
            self.assertEqual(
                g.get_property("Conventions"),
                " ".join(Conventions),
                f"{g.get_property('Conventions')!r}, {Conventions!r}",
            )

        for Conventions in ([other, version],):
            cfdm.write(f, tmpfile0, Conventions=Conventions)
            g = cfdm.read(tmpfile0)[0]
            self.assertEqual(
                g.get_property("Conventions"),
                " ".join([version, other]),
                f"{g.get_property('Conventions')!r}, {Conventions!r}",
            )

    def test_read_write_multiple_geometries(self):
        """Test reading and writing with a mixture of geometry cells."""
        a = []
        for filename in (
            "geometry_1.nc",
            "geometry_2.nc",
            "geometry_3.nc",
            "geometry_4.nc",
            "geometry_interior_ring_2.nc",
            "geometry_interior_ring.nc",
        ):
            a.extend(cfdm.read(filename))

        for n, f in enumerate(a):
            f.set_property("test_id", str(n))

        cfdm.write(a, tmpfile, verbose=1)

        f = cfdm.read(tmpfile, verbose=1)

        self.assertEqual(len(a), len(f))

        for x in a:
            for n, y in enumerate(f[:]):
                if x.equals(y):
                    f.pop(n)
                    break

        self.assertFalse(f)

    def test_read_write_domain(self):
        """Test the reading and writing of domain constucts."""
        f = self.f1
        d = f.domain.copy()

        # 1 domain
        cfdm.write(d, tmpfile)
        e = cfdm.read(tmpfile)
        self.assertTrue(len(e), 10)

        e = cfdm.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 1)
        e = e[0]
        self.assertIsInstance(e, cfdm.Domain)
        self.assertTrue(e.equals(e.copy(), verbose=3))
        self.assertTrue(d.equals(e, verbose=3))
        self.assertTrue(e.equals(d, verbose=3))

        # 1 field and 1 domain
        cfdm.write([f, d], tmpfile)
        g = cfdm.read(tmpfile)
        self.assertTrue(len(g), 1)
        g = g[0]
        self.assertIsInstance(g, cfdm.Field)
        self.assertTrue(g.equals(f, verbose=3))

        e = cfdm.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 1)
        e = e[0]
        self.assertIsInstance(e, cfdm.Domain)

        # 1 field and 2 domains
        cfdm.write([f, d, d], tmpfile)
        g = cfdm.read(tmpfile)
        self.assertTrue(len(g), 1)
        g = g[0]
        self.assertIsInstance(g, cfdm.Field)
        self.assertTrue(g.equals(f, verbose=3))

        e = cfdm.read(tmpfile, domain=True, verbose=1)
        self.assertEqual(len(e), 2)
        self.assertIsInstance(e[0], cfdm.Domain)
        self.assertIsInstance(e[1], cfdm.Domain)
        self.assertTrue(e[0].equals(e[1]))

    def test_write_coordinates(self):
        """Test the `coordinates` keyword argument of `write`."""
        f = self.f0

        cfdm.write(f, tmpfile, coordinates=True)
        g = cfdm.read(tmpfile)

        self.assertEqual(len(g), 1)
        self.assertTrue(g[0].equals(f, verbose=3))

    def test_write_scalar_domain_ancillary(self):
        """Test the writing of a file with a scalar domain ancillary."""
        f = self.f1.copy()

        # Create scalar domain ancillary
        d = f.construct("ncvar%a")
        d.del_data()
        d.set_data(10)
        d.del_bounds()

        key = f.construct_key("ncvar%a")
        f.set_data_axes((), key=key)

        cfdm.write(f, tmpfile)

    def test_write_filename_expansion(self):
        """Test the writing to a file name that requires expansions."""
        f = self.f0
        filename = os.path.join("$PWD", os.path.basename(tmpfile))
        cfdm.write(f, filename)

    def test_read_zero_length_file(self):
        """Test reading a zero length file raises an exception."""
        # Create zero-length file
        tmpfile = tempfile.mkstemp("_test_read_write.nc", dir=os.getcwd())[1]
        tmpfiles.append(tmpfile)
        subprocess.run(f"touch {tmpfile}", shell=True, check=True)

        with self.assertRaises(DatasetTypeError):
            cfdm.read(tmpfile)

    def test_read_subsampled_coordinates(self):
        """Test the reading subsampled coordinates."""
        for i in cfdm.read("subsampled_2.nc"):
            identity = i.identity()
            if identity == "long_name=radiance":
                # Field with subsampled coordinates
                r = i
                continue

            if identity == "latitude":
                # Original latitudes
                olat = i
                continue

            if identity == "longitude":
                # Original longitudes
                olon = i
                continue

        # Check that reconstituted coordinates equal the original
        # coordinates
        rlat = r.construct("latitude").data.array
        rlon = r.construct("longitude").data.array

        self.assertTrue(rlat.shape, olat.shape)
        self.assertTrue(rlon.shape, olon.shape)

        self.assertTrue(
            np.allclose(rlat, olat.data.array, atol=5.4e-06, rtol=0)
        )
        self.assertTrue(
            np.allclose(rlon, olon.data.array, atol=2.2e-05, rtol=0)
        )

    def test_read_original_filenames(self):
        """Test the setting of original file names."""
        f = cfdm.read(self.filename)[0]
        x = f.dimension_coordinate("grid_longitude")

        for a in (x, f):
            self.assertEqual(
                a.get_original_filenames(), set([cfdm.abspath(self.filename)])
            )

        # Two original files
        parent_file = "parent.nc"
        external_file = "external.nc"
        f = cfdm.read(parent_file, external=external_file)[0]
        self.assertEqual(
            f.get_original_filenames(),
            set((cfdm.abspath(parent_file), cfdm.abspath(external_file))),
        )

    def test_write_omit_data(self):
        """Test the `omit_data` parameter to `write`."""
        f = self.f1
        cfdm.write(f, tmpfile)

        cfdm.write(f, tmpfile, omit_data="all")
        g = cfdm.read(tmpfile)
        self.assertEqual(len(g), 1)
        g = g[0]

        # Check that the data are missing
        self.assertFalse(np.ma.count(g.array))
        self.assertFalse(np.ma.count(g.construct("grid_latitude").array))

        # Check that a dump works
        g.dump(display=False)

        cfdm.write(f, tmpfile, omit_data=("field", "dimension_coordinate"))
        g = cfdm.read(tmpfile)[0]

        # Check that only the field and dimension coordinate data are
        # missing
        self.assertFalse(np.ma.count(g.array))
        self.assertFalse(np.ma.count(g.construct("grid_latitude").array))
        self.assertTrue(np.ma.count(g.construct("latitude").array))

        cfdm.write(f, tmpfile, omit_data="field")
        g = cfdm.read(tmpfile)[0]

        # Check that only the field data are missing
        self.assertFalse(np.ma.count(g.array))
        self.assertTrue(np.ma.count(g.construct("grid_latitude").array))

    def test_read_write_domain_ancillary(self):
        """Test when domain ancillary equals dimension coordinate."""
        f = self.f1

        # Check the domain ancillary does indeed equal the dimension
        # coordinate
        self.assertTrue(
            f.domain_ancillary("atmosphere_hybrid_height_coordinate").equals(
                f.dimension_coordinate("atmosphere_hybrid_height_coordinate"),
                ignore_type=True,
            )
        )

        cfdm.write(f, tmpfile)
        g = cfdm.read(tmpfile)
        self.assertEqual(len(g), 1)
        g = g[0]
        self.assertTrue(f.equals(g))

        nc = netCDF4.Dataset(tmpfile, "r")

        z = nc.variables["atmosphere_hybrid_height_coordinate"]
        # In the following test we are checking that it is not 'a: a
        # b: b orog: surface_altitude'
        self.assertEqual(
            z.getncattr("formula_terms"),
            "a: atmosphere_hybrid_height_coordinate b: b orog: surface_altitude",
        )

        zb = nc.variables["atmosphere_hybrid_height_coordinate_bounds"]
        # In the following test we are checking that it is not 'a:
        # a_bounds b: b orog: surface_altitude'
        self.assertEqual(
            zb.getncattr("formula_terms"),
            "a: atmosphere_hybrid_height_coordinate_bounds b: b_bounds orog: surface_altitude",
        )

        nc.close()

        # Test the full round trip
        cfdm.write(g, tmpfile0)
        h = cfdm.read(tmpfile0)
        self.assertEqual(len(h), 1)
        self.assertTrue(f.equals(h[0]))

    def test_read_url(self):
        """Test reading urls."""
        return  # skip flaky test until it is made robust
        for scheme in ("http", "https"):
            remote = f"{scheme}://psl.noaa.gov/thredds/dodsC/Datasets/cru/crutem5/Monthlies/air.mon.anom.nobs.nc"
            # Check that cfdm can access it
            f = cfdm.read(remote)
            self.assertEqual(len(f), 1)

    def test_write_parametric_Z_coordinate(self):
        """Test write of parametric Z coordinate."""
        # Thes write when a parametric Z dimension coordinate does not
        # have a compute_standard_name attribute
        f = self.f1.copy()
        f.coordinate("atmosphere_hybrid_height_coordinate").del_property(
            "computed_standard_name", None
        )
        cfdm.write(f, tmpfile)

    def test_write_dataset_chunks(self):
        """Test the 'dataset_chunks' parameter to `cfdm.write`."""
        f = cfdm.example_field(5)
        f.nc_set_variable("data")

        # Good dataset_chunks values
        for dataset_chunks, chunking in zip(
            ("4MiB", "8KiB", "5000", 314.159, 1, "contiguous"),
            (
                [118, 5, 8],
                [25, 5, 8],
                [15, 5, 8],
                [3, 3, 3],
                [1, 1, 1],
                "contiguous",
            ),
        ):
            cfdm.write(f, tmpfile, dataset_chunks=dataset_chunks)
            nc = netCDF4.Dataset(tmpfile, "r")
            self.assertEqual(nc.variables["data"].chunking(), chunking)
            nc.close()

        # Bad dataset_chunks values
        for dataset_chunks in ("bad_value", None):
            with self.assertRaises(ValueError):
                cfdm.write(f, tmpfile, dataset_chunks=dataset_chunks)

        # Check that user-set chunks are not overridden
        for chunking in ([5, 4, 3], "contiguous"):
            f.nc_set_hdf5_chunksizes(chunking)
            for dataset_chunks in ("4MiB", "contiguous"):
                cfdm.write(f, tmpfile, dataset_chunks=dataset_chunks)
                nc = netCDF4.Dataset(tmpfile, "r")
                self.assertEqual(nc.variables["data"].chunking(), chunking)
                nc.close()

        f.nc_set_hdf5_chunksizes("120 B")
        for dataset_chunks in ("contiguous", "4MiB"):
            cfdm.write(f, tmpfile, dataset_chunks=dataset_chunks)
            nc = netCDF4.Dataset(tmpfile, "r")
            self.assertEqual(nc.variables["data"].chunking(), [2, 2, 2])
            nc.close()

        # store_dataset_chunks
        f = cfdm.read(tmpfile)[0]
        self.assertEqual(f.nc_hdf5_chunksizes(), (2, 2, 2))

        f = cfdm.read(tmpfile, store_dataset_chunks=False)[0]
        self.assertIsNone(f.nc_hdf5_chunksizes())

        # Scalar data is written contiguously
        f = self.f0
        f = f[0, 0].squeeze()
        cfdm.write(f, tmpfile)
        nc = netCDF4.Dataset(tmpfile, "r")
        self.assertEqual(nc.variables["q"].chunking(), "contiguous")
        nc.close()

    def test_read_dask_chunks(self):
        """Test the 'dask_chunks' keyword of cfdm.read."""
        f = self.f0.copy()
        f.coordinate("latitude").axis = "Y"
        cfdm.write(f, tmpfile)

        # Dictionary
        f = cfdm.read(tmpfile, dask_chunks={})[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cfdm.read(tmpfile, dask_chunks={"foo": 2, "bar": 3})[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cfdm.read(tmpfile, dask_chunks={"ncdim%lon": 3})[0]
        self.assertEqual(f.data.chunks, ((5,), (3, 3, 2)))

        f = cfdm.read(tmpfile, dask_chunks={"longitude": 6, "Y": "150B"})[0]
        self.assertEqual(f.data.chunks, ((5,), (6, 2)))

        y = f.construct("latitude")
        self.assertEqual(y.data.chunks, ((5,),))

        # -1, None
        f = cfdm.read(tmpfile, dask_chunks=-1)[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        f = cfdm.read(tmpfile, dask_chunks=None)[0]
        self.assertEqual(f.data.chunks, ((5,), (8,)))

        # Positive integer
        f = cfdm.read(tmpfile, dask_chunks=3)[0]
        self.assertEqual(f.data.chunks, ((3, 2), (3, 3, 2)))

        y = f.construct("latitude")
        self.assertEqual(y.data.chunks, ((3, 2),))

        f = cfdm.read(tmpfile, dask_chunks="150B")[0]
        self.assertEqual(f.data.chunks, ((4, 1), (4, 4)))

        # storage-exact
        f = cfdm.example_field(2)
        f.data.nc_set_hdf5_chunksizes([7, 5, 4])
        cfdm.write(f, tmpfile)
        g = cfdm.read(tmpfile, dask_chunks="storage-exact")[0]
        self.assertEqual(g.data.chunks, ((7, 7, 7, 7, 7, 1), (5,), (4, 4)))

        # storage-aligned (the default)
        g = cfdm.read(tmpfile, dask_chunks="storage-aligned")[0]
        self.assertEqual(g.data.chunks, ((35, 1), (5,), (8,)))

        g = cfdm.read(tmpfile)[0]
        self.assertEqual(g.data.chunks, ((35, 1), (5,), (8,)))

        with cfdm.chunksize(50000000):
            g = cfdm.read(tmpfile)[0]
            self.assertEqual(g.data.chunks, ((35, 1), (5,), (8,)))

        with cfdm.chunksize(5000):
            g = cfdm.read(tmpfile)[0]
            self.assertEqual(g.data.chunks, ((14, 14, 8), (5,), (8,)))

        with cfdm.chunksize(500):
            g = cfdm.read(tmpfile)[0]
            self.assertEqual(g.data.chunks, ((7, 7, 7, 7, 7, 1), (5,), (4, 4)))

    def test_read_to_memory(self):
        """Test the 'to_memory' parameter to cfdm.read."""
        f = self.f0
        cfdm.write(f, tmpfile)

        f = cfdm.read(tmpfile)[0]
        for d in (f.data.todict(), f.coordinate("longitude").data.todict()):
            on_disk = False
            for v in d.values():
                if isinstance(v, cfdm.H5netcdfArray):
                    on_disk = True

            self.assertTrue(on_disk)

        for to_memory in ("all", ("field", "dimension_coordinate")):
            f = cfdm.read(tmpfile, to_memory=to_memory)[0]
            for d in (
                f.data.todict(),
                f.coordinate("longitude").data.todict(),
            ):
                in_memory = False
                for v in d.values():
                    if isinstance(v, np.ndarray):
                        in_memory = True

                self.assertTrue(in_memory)

        for to_memory in ("metadata", "dimension_coordinate"):
            f = cfdm.read(tmpfile, to_memory=to_memory)[0]
            for i, d in enumerate(
                (
                    f.coordinate("longitude").data.todict(),
                    f.data.todict(),
                )
            ):
                in_memory = False
                for v in d.values():
                    if isinstance(v, np.ndarray):
                        in_memory = True

                if not i:
                    # Metadata
                    self.assertTrue(in_memory)
                else:
                    # Field
                    self.assertFalse(in_memory)

    def test_read_file_type(self):
        """Test the cfdm.read 'file_type' keyword."""
        # netCDF file
        for file_type in (
            None,
            "netCDF",
            ("netCDF",),
            ("netCDF", "CDL"),
            ("netCDF", "CDL", "bad value"),
        ):
            f = cfdm.read(self.filename, file_type=file_type)
            self.assertEqual(len(f), 1)

        for file_type in ("CDL", "bad value"):
            f = cfdm.read(self.filename, file_type=file_type)
            self.assertEqual(len(f), 0)

        # CDL file
        subprocess.run(
            " ".join(["ncdump", self.filename, ">", tmpfile]),
            shell=True,
            check=True,
        )
        for file_type in (
            None,
            "CDL",
            ("CDL",),
            ("netCDF", "CDL"),
            ("netCDF", "CDL", "bad value"),
        ):
            f = cfdm.read(tmpfile, file_type=file_type)
            self.assertEqual(len(f), 1)

        for file_type in ("netCDF", "bad value"):
            f = cfdm.read(tmpfile, file_type=file_type)
            self.assertEqual(len(f), 0)

        # Not a netCDF or CDL file
        with self.assertRaises(DatasetTypeError):
            f = cfdm.read("test_read_write.py")

        for file_type in ("netCDF", "CDL", "bad value"):
            f = cfdm.read("test_read_write.py", file_type=file_type)
            self.assertEqual(len(f), 0)


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print("")
    unittest.main(verbosity=2)



"""Define docstring substitutions.

Text to be replaced is specified as a key in the returned dictionary,
with the replacement text defined by the corresponding value.

Special docstring substitutions, as defined by a class's
`_docstring_special_substitutions` method, may be used in the
replacement text, and will be substituted as usual.

Replacement text may not contain other non-special substitutions.

Keys must be a `str` or `re.Pattern` object:

* If a key is a `str` then the corresponding value must be a string.

* If a key is a `re.Pattern` object then the corresponding value must
  be a string or a callable, as accepted by the `re.Pattern.sub`
  method.

.. versionaddedd:: (cfdm) 1.8.7.0

"""

_docstring_substitution_definitions = {
    # ----------------------------------------------------------------
    # General substitutions (not indent-dependent)
    # ----------------------------------------------------------------
    #
    # ----------------------------------------------------------------
    # Class description substitutions (1 level of indentation)
    # ----------------------------------------------------------------
    # netCDF variable
    "{{netCDF variable}}": """The netCDF variable name may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_del_variable`, and
    `nc_has_variable` methods.

    The netCDF variable group structure may be accessed with the
    `nc_set_variable`, `nc_get_variable`, `nc_variable_groups`,
    `nc_clear_variable_groups`, and `nc_set_variable_groups`
    methods.""",
    # netCDF global attributes
    "{{netCDF global attributes}}": """The selection of properties to be written as netCDF global
    attributes may be accessed with the `nc_global_attributes`,
    `nc_clear_global_attributes`, and `nc_set_global_attribute`
    methods.""",
    # netCDF group attributes
    "{{netCDF group attributes}}": """The netCDF group attributes may be accessed with the
    `nc_group_attributes`, `nc_clear_group_attributes`,
    `nc_set_group_attribute`, and `nc_set_group_attributes` methods.""",
    # netCDF geometry group
    "{{netCDF geometry group}}": """The netCDF geometry variable name and group structure may be
    accessed with the `nc_set_geometry_variable`,
    `nc_get_geometry_variable`, `nc_geometry_variable_groups`,
    `nc_clear_variable_groups`, and `nc_set_geometry_variable_groups`
    methods.""",
    # netCDF UGRID node coordinate
    "{{netCDF UGRID node coordinate}}": """The netCDF UGRID node coordinate variable name may be accessed
    with the `nc_set_node_coordinate_variable`,
    `nc_get_node_coordinate_variable`,
    `nc_del_node_coordinate_variable`, and
    `nc_has_node_coordinate_variable` methods.

    The netCDF UGRID node coordinate variable group structure may be
    accessed with the `nc_set_node_coordinate_variable`,
    `nc_get_node_coordinate_variable`,
    `nc_variable_node_coordinate_groups`,
    `nc_clear_node_coordinate_variable_groups`, and
    `nc_set_node_coordinate_variable_groups` methods.""",
    # netCDF HDF5 chunks
    "{{netCDF HDF5 chunks}}": """The netCDF4 HDF5 chunks may be accessed with the
    `nc_hdf5_chunksizes`, `nc_set_hdf5_chunksizes`, and
    `nc_clear_hdf5_chunksizes` methods.""",
    # ----------------------------------------------------------------
    # Method description substitutions (2 levels of indentation)
    # ----------------------------------------------------------------
    # equals tolerance
    "{{equals tolerance}}": """Two real numbers ``x`` and ``y`` are considered equal if
        ``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on
        absolute differences) and ``rtol`` (the tolerance on relative
        differences) are positive, typically very small numbers. See
        the *atol* and *rtol* parameters.""",
    # equals compression
    "{{equals compression}}": """Any compression is ignored by default, with only the arrays in
        their uncompressed forms being compared. See the
        *ignore_compression* parameter.""",
    # equals netCDF
    "{{equals netCDF}}": """NetCDF elements, such as netCDF variable and dimension names,
        do not constitute part of the CF data model and so are not
        checked.""",
    # netcdf global
    "{{netcdf global}}": """When multiple field or domain constructs are being written to
        the same file, it is only possible to create a netCDF global
        attribute from a property that has identical values for each
        construct. If any field or domain construct's property has a
        different value then the property will not be written as a
        netCDF global attribute, even if it has been selected as such,
        but will appear instead as attributes on the netCDF data
        variables corresponding to each construct.

        The standard description-of-file-contents properties are
        always written as netCDF global attributes, if possible, so
        selecting them is optional.""",
    # unique construct
    "{{unique construct}}": """If zero or two or more constructs are selected then an
        exception is raised, or the *default* parameter is returned.""",
    # original filenames
    "{{original filenames}}": """The original files are those that contain some or all of the
        data and metadata when it was first instantiated, and are
        necessary (but perhaps not sufficient) to recreate the
        `{{class}}` should the need arise. The `{{package}}.read`
        function automatically records the original file names on all
        data that it creates.

        The original files of any constituent components are also
        included.

        In-place changes to the `{{class}}` will not generally change
        the collection of original files. However if the `{{class}}`
        was produced by combining other objects that also store their
        original file names, then the returned files will be the
        collection of original files from all contributing sources.""",
    # read external
    "{{read external: (sequence of) `str`, optional}}": """external: (sequence of) `str`, optional
            Read external variables (i.e. variables which are named by
            attributes, but are not present, in the parent file given
            by the *filename* parameter) from the given external
            files. Ignored if the parent file does not contain a
            global ``external_variables`` attribute. Multiple external
            files may be provided, which are searched in random order
            for the required external variables.

            If an external variable is not found in any external
            files, or is found in multiple external files, then the
            relevant metadata construct is still created, but without
            any metadata or data. In this case the construct's
            `!is_external` method will return `True`.

            *Parameter example:*
              ``external='cell_measure.nc'``

            *Parameter example:*
              ``external=['cell_measure.nc']``

            *Parameter example:*
              ``external=('cell_measure_A.nc', 'cell_measure_O.nc')``""",
    # read extra
    "{{read extra: (sequence of) `str`, optional}}": """extra: (sequence of) `str`, optional
            Create extra, independent fields from netCDF variables
            that correspond to particular types of metadata constructs.
            Ignored if *domain* is True.

            The *extra* parameter may be one, or a sequence, of:

            ==========================  ===============================
            *extra*                     Metadata constructs
            ==========================  ===============================
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ==========================  ===============================

            *Parameter example:*
              To create fields from auxiliary coordinate constructs:
              ``extra='auxiliary_coordinate'`` or
              ``extra=['auxiliary_coordinate']``.

            *Parameter example:*
              To create fields from domain ancillary and cell measure
              constructs: ``extra=['domain_ancillary',
              'cell_measure']``.

            An extra field construct created via the *extra* parameter
            will have a domain limited to that which can be inferred
            from the corresponding netCDF variable, but without the
            connections that are defined by the parent netCDF data
            variable. It is possible to create independent fields from
            metadata constructs that do incorporate as much of the
            parent field construct's domain as possible by using the
            `~{{package}}.Field.convert` method of a returned field
            construct, instead of setting the *extra* parameter.""",
    # read verbose
    "{{read verbose: `int` or `str` or `None`, optional}}": """verbose: `int` or `str` or `None`, optional
            If an integer from ``-1`` to ``3``, or an equivalent string
            equal ignoring case to one of:

            * ``'DISABLE'`` (``0``)
            * ``'WARNING'`` (``1``)
            * ``'INFO'`` (``2``)
            * ``'DETAIL'`` (``3``)
            * ``'DEBUG'`` (``-1``)

            set for the duration of the method call only as the
            minimum cut-off for the verboseness level of displayed
            output (log) messages, regardless of the
            globally-configured `{{package}}.log_level`.  Note that
            increasing numerical value corresponds to increasing
            verbosity, with the exception of ``-1`` as a special case
            of maximal and extreme verbosity.

            Otherwise, if `None` (the default value), output messages
            will be shown according to the value of the
            `{{package}}.log_level` setting.

            Overall, the higher a non-negative integer or equivalent string
            that is set (up to a maximum of ``3``/``'DETAIL'``) for
            increasing verbosity, the more description that is printed to
            convey how the contents of the netCDF file were parsed and
            mapped to CF data model constructs.""",
    # read warnings
    "{{read warnings: `bool`, optional}}": """warnings: `bool`, optional
            If True then print warnings when an output field construct
            is incomplete due to structural non-compliance of the
            dataset. By default such warnings are not displayed.""",
    # read warn_valid
    "{{read warn_valid: `bool`, optional}}": """warn_valid: `bool`, optional
            If True then print a warning for the presence of
            ``valid_min``, ``valid_max`` or ``valid_range`` properties
            on field constructs and metadata constructs that have
            data. By default no such warning is issued.

            "Out-of-range" data values in the file, as defined by any
            of these properties, are automatically masked by default,
            which may not be as intended. See the *mask* parameter for
            turning off all automatic masking.""",
    # read mask
    "{{read mask: `bool`, optional}}": """mask: `bool`, optional
            If True (the default) then mask by convention the data of
            field and metadata constructs.

            A netCDF array is masked depending on the values of any of
            the netCDF attributes ``_FillValue``, ``missing_value``,
            ``_Unsigned``, ``valid_min``, ``valid_max``, and
            ``valid_range``.""",
    # read unpack
    "{{read unpack: `bool`}}": """unpack: `bool`
            If True, the default, then unpack arrays by convention
            when the data is read from disk.

            Unpacking is determined by netCDF conventions for the
            following variable attributes: ``add_offset``,
            ``scale_factor``, and ``_Unsigned``.""",
    # read domain
    "{{read domain: `bool`, optional}}": """domain: `bool`, optional
            If True then return only the domain constructs that are
            explicitly defined by CF-netCDF domain variables, ignoring
            all CF-netCDF data variables. By default only the field
            constructs defined by CF-netCDF data variables are
            returned.

            CF-netCDF domain variables are only defined from CF-1.9,
            so older datasets automatically contain no CF-netCDF
            domain variables.

            The unique domain constructs of the dataset are
            found with the `{{package}}.unique_constructs`
            function. For example::

               >>> d = {{package}}.read('file.nc', domain=True)
               >>> ud = {{package}}.unique_constructs(d)
               >>> f = {{package}}.read('file.nc')
               >>> ufd = {{package}}.unique_constructs(x.domain for x in f)""",
    # read netcdf_backend
    "{{read netcdf_backend: `None` or (sequence of) `str`, optional}": """netcdf_backend: `None` or (sequence of) `str`, optional
            Specify which library, or libraries, to use for opening
            and reading netCDF files. By default, or if `None`, then
            the first one of `h5netcdf` and `netCDF4` to successfully
            open the netCDF file is used. The libraries will be used
            in the order given, until a file is successfully
            opened.""",
    # read  storage_options
    "{{read storage_options: `dict` or `None`, optional}}": """storage_options: `dict` or `None`, optional
            Pass parameters to the backend file system driver, such as
            username, password, server, port, etc. How the storage
            options are interpreted depends on the location of the
            file:

            * **Local File System**: Storage options are ignored for
              local files.

            * **HTTP(S)**: Storage options are ignored for files
              available across the network via OPeNDAP.

            * **S3-compatible services**: The backend used is `s3fs`,
              and the storage options are used to initialise an
              `s3fs.S3FileSystem` file system object. By default, or
              if `None`, then *storage_options* is taken as ``{}``.

              If the ``'endpoint_url'`` key is not in
              *storage_options*, nor in a dictionary defined by the
              ``'client_kwargs'`` key (both of which are the case when
              *storage_options* is `None`), then one will be
              automatically inserted for accessing an S3 file. For
              instance, with a file name of
              ``'s3://store/data/file.nc'``, an ``'endpoint_url'`` key
              with value ``'https://store'`` would be created. To
              disable this, set the ``'endpoint_url'`` key to `None`.

              *Parameter example:*
                For a file name of ``'s3://store/data/file.nc'``, the
                following are equivalent: ``None``, ``{}``,
                ``{'endpoint_url': 'https://store'}``, and
                ``{'client_kwargs': {'endpoint_url':
                'https://store'}}``

              *Parameter example:*
                ``{'key': 'scaleway-api-key...', 'secret':
                'scaleway-secretkey...', 'endpoint_url':
                'https://s3.fr-par.scw.cloud', 'client_kwargs':
                {'region_name': 'fr-par'}}``""",
    # read cache
    "{{read cache: `bool`, optional}}": """cache: `bool`, optional
            If True, the default, then cache the first and last array
            elements of metadata constructs (not field constructs) for
            fast future access. In addition, the second and
            penultimate array elements will be cached from coordinate
            bounds when there are two bounds per cell. For remote
            data, setting *cache* to False may speed up the parsing of
            the file.""",
    # read dask_chunks
    "{{read dask_chunks: `str`, `int`, `None`, or `dict`, optional}}": """dask_chunks: `str`, `int`, `None`, or `dict`, optional
            Specify the Dask chunking for data. May be one of the
            following:

            * ``'storage-aligned'``

              This is the default. The Dask chunk size in bytes will
              be as close as possible the size given by
              `{{package}}.chunksize`, favouring square-like chunk
              shapes, with the added restriction that the entirety of
              each storage chunk lies within exactly one Dask
              chunk. This strategy is general the most performant, as
              it ensures that when accessing the data, each storage
              chunk is read from disk at most once (as opposed to once
              per Dask chunk in which it lies).

              For instance, consider a file variable that has an array
              of 64-bit floats with shape (400, 300, 60) and a storage
              chunk shape of (100, 5, 60). This has an overall size of
              54.93 MiB, partitioned into 240 storage chunks each of
              size 100*5*60*8 bytes = 0.23 MiB. Then:

              * If `{{package}}.chunksize` returns 134217728 (i.e. 128
                MiB), then the storage-aligned Dask chunks will have
                shape (400, 300, 60), giving 1 Dask chunk with size of
                54.93 MiB (compare with a Dask chunk shape of (400,
                300, 60) and size 54.93 MiB when *dask_chunks* is
                ``'auto'``.)

              * If `{{package}}.chunksize` returns 33554432 (i.e. 32
                MiB), then the storage-aligned Dask chunks will have
                shape (200, 260, 60), giving 4 Dask chunks with a
                maximum size of 23.80 MiB (compare with a Dask chunk
                shape of (264, 264, 60) and maximum size 31.90 MiB
                when *dask_chunks* is ``'auto'``.)

              * If `{{package}}.chunksize` returns 4194304 (i.e. 4
                MiB), then the storage-aligned Dask chunks will have
                shape (100, 85, 60), giving 16 Dask chunks with a
                maximum size of 3.89 MiB (compare with a Dask chunk
                shape of (93, 93, 60) and maximum size 3.96 MiB when
                *dask_chunks* is ``'auto'``.)

              There are, however, some occasions when, for particular
              data arrays in the file, the ``'auto'`` option will
              automatically be used instead of storage-aligned Dask
              chunks. This occurs when:

              * The data array in the file is stored contiguously.

              * The data array in the file is compressed by convention
                (e.g. ragged array representations, compression by
                gathering, subsampled coordinates, etc.). In this case
                the Dask chunks are for the uncompressed data, and so
                cannot be aligned with the storage chunks of the
                compressed array in the file.

            * ``'storage-exact'``

              Each Dask chunk will contain exactly one storage chunk
              and each storage chunk will lie within exactly one Dask
              chunk.

              For instance, consider a file variable that has an array
              of 64-bit floats with shape (400, 300, 60) and a storage
              chunk shape of (100, 5, 60). This has an overall size of
              54.93 MiB, partitioned into 240 storage chunks each of
              size 100*5*60*8 bytes = 0.23 MiB. The corresponding
              storage-exact Dask chunks will also have shape (100, 5,
              60), giving 240 Dask chunks with a maximum size of 0.23
              MiB.

              There are, however, some occasions when, for particular
              data arrays in the file, the ``'auto'`` option will
              automatically be used instead of storage-exact Dask
              chunks. This occurs when:

              * The data array in the file is stored contiguously.

              * The data array in the file is compressed by convention
                (e.g. ragged array representations, compression by
                gathering, subsampled coordinates, etc.). In this case
                the Dask chunks are for the uncompressed data, and so
                cannot be aligned with the storage chunks of the
                compressed array in the file.

            * ``auto``

              The Dask chunk size in bytes will be as close as
              possible to the size given by `{{package}}.chunksize`,
              favouring square-like chunk shapes. This may give
              similar Dask chunk shapes as the ``'storage-aligned'``
              option, but without the guarantee that each storage
              chunk will lie within exactly one Dask chunk.

            * A byte-size given by a `str`

              The Dask chunk size in bytes will be as close as
              possible to the given byte-size, favouring square-like
              chunk shapes. Any string value, accepted by the *chunks*
              parameter of the `dask.array.from_array` function is
              permitted.

              *Example:*
                A Dask chunksize of 2 MiB may be specified as
                ``'2097152'`` or ``'2 MiB'``.

            * `-1` or `None`

              There is no Dask chunking, i.e. every data array has one
              Dask chunk regardless of its size.

            * Positive `int`

              Every dimension of all Dask chunks has this number of
              elements.

              *Example:*
                For 3-dimensional data, *dask_chunks* of `10` will
                give Dask chunks with shape (10, 10, 10).

            * `dict`

              Each of dictionary key identifies a file dimension, with
              a value that defines the Dask chunking for that
              dimension whenever it is spanned by a data array. A file
              dimension is identified in one of three ways:

              1. the netCDF dimension name, preceded by ``ncdim%``
                (e.g. ``'ncdim%lat'``);

              2. the value of the "standard name" attribute of a
                 CF-netCDF coordinate variable that spans the
                 dimension (e.g. ``'latitude'``);

              3. the value of the "axis" attribute of a CF-netCDF
                 coordinate variable that spans the dimension
                 (e.g. ``'Y'``).

              The dictionary values may be a byte-size string,
              ``'auto'``, `int` or `None`, with the same meanings as
              those types for the *dask_chunks* parameter itself, but
              applying only to the specified dimension. In addition, a
              dictionary value may be a `tuple` or `list` of integers
              that sum to the dimension size.

              Not specifying a file dimension in the dictionary is
              equivalent to it being defined with a value of
              ``'auto'``.

              *Example:*
                ``{'T': '0.5 MiB', 'Z': 'auto', 'Y': [36, 37], 'X':
                None}``

              *Example:*
                If a netCDF file contains dimensions ``time``, ``z``,
                ``lat`` and ``lon``, then ``{'ncdim%time': 12,
                'ncdim%lat', None, 'ncdim%lon': None}`` will ensure
                that, for all applicable data arrays, all ``time``
                axes have a `dask` chunksize of 12; all ``lat`` and
                ``lon`` axes are not `dask` chunked; and all ``z``
                axes are `dask` chunked to comply as closely as
                possible with the default `dask` chunk size.

                If the netCDF file also contains a ``time`` coordinate
                variable with a "standard_name" attribute of
                ``'time'`` or "axis" attribute of ``'T'``, then the
                same `dask` chunking could be specified with either
                ``{'time': 12, 'ncdim%lat', None, 'ncdim%lon': None}``
                or ``{'T': 12, 'ncdim%lat', None, 'ncdim%lon':
                None}``.""",
    # read store_hdf5_chunks
    "{{read store_dataset_chunks: `bool`, optional}}": """store_dataset_chunks: `bool`, optional
            If True (the default) then store the dataset chunking
            strategy for each returned data array. The dataset
            chunking strategy is then accessible via an object's
            `nc_hdf5_chunksizes` method. When the dataset chunking
            strategy is stored, it will be used when the data is
            written to a new netCDF4 file with `{{package}}.write`
            (unless the strategy is modified prior to writing).

            If False, or if the dataset format does not support
            chunking, then no dataset chunking strategy is stored.
            (i.e. an `nc_hdf5_chunksizes` method will return `None`
            for all `Data` objects). In this case, when the data is
            written to a new netCDF4 file, the dataset chunking
            strategy will be determined by `{{package}}.write`.

            See the `{{package}}.write` *hdf5_chunks* parameter for
            details on how the dataset chunking strategy is determined
            at the time of writing.""",
    # read cfa
    "{{read cfa: `dict`, optional}}": """cfa: `dict`, optional
            Configure the reading of CF-netCDF aggregation files.

            The dictionary may have any subset of the following
            key/value pairs that supplement or override the
            information read from the file:

            * ``'replace_directory'``: `dict`

              A dictionary whose key/value pairs define modifications
              to be applied to the directories of the fragment file
              locations. The dictionary comprises keyword arguments to
              the {{package}}.Data.replace_directory` method, which is
              used to make the the changes. The aggregation file being
              read is unaltered. An empty dictionary results in no
              modifications.

              *Example:*
                Replace a leading ``data/model`` with ``home``,
                wherever it occurs: ``{'replace_directory': {'old':
                'data/model', 'new': 'home'}}``

              *Example:*
                Normalise all file locations and replace a leading
                ``/archive`` with ``/data/obs``, wherever it occurs:
                ``{'replace_directory': {'old': '/archive', 'new':
                '/data/obs', 'normalise': True}}``

              *Example:*
                Normalise all file locations and remove a leading
                ``/data`, wherever it occurs: ``{'replace_directory':
                {'old': '/data', 'normalise': True}}``.""",
    # read cfa_write
    "{{read cfa_write: (sequence of) `str`, optional}}": """cfa_write: (sequence of) `str`, optional
            Register the intention for named construct types to be
            subsequently written as CF-netCDF aggregation variables.

            This makes no difference to the logical content of any
            construct, but ensures that the data of each of specified
            construct types will have only one Dask chunk, regardless
            of the setting of *dask_chunks*, which is a requirement
            for the creation CF-netCDF aggregation variables.

            The *cfa_write* parameter may be one, or a sequence, of:

            ==========================  ===============================
            *cfa_write*                 Construct types
            ==========================  ===============================
            ``'field'``                 Field constructs
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ``'domain_topology'``       Domain topology constructs
            ``'cell_connectivity'``     Cell connectivity constructs
            ``'all'``                   All constructs
            ==========================  ===============================

            .. note:: If the *dask_chunks* parameter is set to `None`
                      or ``-1`` then the data of all constructs will
                      already have only one Dask chunk, so in this
                      case setting *cfa_write* will have no further
                      effect.

            *Example:*
              To register field constructs to be written as CF-netCDF
              aggregation variables: ``cfa_write='field'`` or
              ``cfa_write=['field']``.

            *Example:*
              To register field and auxiliary coordinate constructs to
              be written as CF-netCDF aggregation variables:
              ``cfa_write=['field', 'auxiliary_coordinate']``.""",
    # read to_memory
    "{{read to_memory: (sequence of) `str`, optional}}": """to_memory: (sequence of) `str`, optional
            Read all data arrays of the named construct types into
            memory. By default, lazy loading is employed for all data
            arrays.

            The *to_memory* parameter may be one, or a sequence, of:

            ==========================  ===============================
            *to_memory*                 Construct types
            ==========================  ===============================
            ``'all'``                   All constructs
            ``'metadata'``              All metadata constructs (i.e.
                                        all constructs except Field
                                        constructs)

            ``'field'``                 Field constructs
            ``'field_ancillary'``       Field ancillary constructs
            ``'domain_ancillary'``      Domain ancillary constructs
            ``'dimension_coordinate'``  Dimension coordinate constructs
            ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
            ``'cell_measure'``          Cell measure constructs
            ``'domain_topology'``       Domain topology constructs
            ``'cell_connectivity'``     Cell connectivity constructs
            ==========================  ===============================

            *Example:*
              To read field construct data arrays into memory:
              ``to_memory='field'`` or ``to_memory=['field']``.

            *Example:*
              To read field and auxiliary coordinate construct data
              arrays into memory: ``to_memory=['field',
              'auxiliary_coordinate']``.""",
    # read squeeze
    "{{read squeeze: `bool`, optional}}": """squeeze: `bool`, optional
            If True then remove all size 1 dimensions from field
            construct data arrays, regardless of how the data are
            stored in the dataset. If False (the default) then the
            presence or not of size 1 dimensions is determined by how
            the data are stored in its dataset.""",
    # read unsqueeze
    "{{read unsqueeze: `bool`, optional}}": """unsqueeze: `bool`, optional
            If True then ensure that field construct data arrays span
            all of the size 1 dimensions, regardless of how the data
            are stored in the dataset. If False (the default) then the
            presence or not of size 1 dimensions is determined by how
            the data are stored in its dataset.""",
    # read file_type
    "{{read file_type: `None` or (sequence of) `str`, optional}}": """file_type: `None` or (sequence of) `str`, optional
             Only read files of the given type(s). All other file
             types are ignored. If `None` (the default) then files of
             any valid type are read. If there are no files of the
             given type(s), or *file_type* is empty sequence, then an
             empty list is returned.""",
    # read ignore_unknown_type
    "{{read ignore_unknown_type: `bool`, optional}}": """ignore_unknown_type: `bool`, optional
             If True then ignore any file which does not have one of
             the valid types specified by the *file_type*
             parameter. If False (the default) then attempting to read
             a file with an unrecognised type will result in an
             error.""",
    # persist
    "{{persist description}}": """Persisting turns an underlying lazy dask array into an
        equivalent chunked dask array, but now with the results fully
        computed and in memory. This can avoid the expense of
        re-reading the data from disk, or re-computing it, when the
        data is accessed on multiple occasions.""",
    # ----------------------------------------------------------------
    # Method description substitutions (3 levels of indentation)
    # ------------------------1----------------------------------------
    # atol: number, optional
    "{{atol: number, optional}}": """atol: number, optional
                The tolerance on absolute differences between real
                numbers. The default value is set by the
                `{{package}}.atol` function.""",
    # data_name: `str`, optional
    "{{data_name: `str`, optional}}": """data_name: `str`, optional
                The name of the construct's `Data` instance created by
                the returned commands.

                *Parameter example:*
                  ``name='data1'``""",
    # header: `bool`, optional
    "{{header: `bool`, optional}}": """header: `bool`, optional
                If True (the default) output a comment describing the
                components. If False no such comment is returned.""",
    # ignore_compression: `bool`, optional
    "{{ignore_compression: `bool`, optional}}": """ignore_compression: `bool`, optional
                If False then the compression type and, if applicable,
                the underlying compressed arrays must be the same, as
                well as the arrays in their uncompressed forms. By
                default only the the arrays in their uncompressed
                forms are compared.""",
    # ignore_data_type: `bool`, optional
    "{{ignore_data_type: `bool`, optional}}": """ignore_data_type: `bool`, optional
                If True then ignore the data types in all numerical
                comparisons. By default different numerical data types
                imply inequality, regardless of whether the elements
                are within the tolerance for equality.""",
    # ignore_fill_value
    "{{ignore_fill_value: `bool`, optional}}": """ignore_fill_value: `bool`, optional
                If True then all ``_FillValue`` and ``missing_value``
                properties are omitted from the comparison.""",
    # ignore_properties
    "{{ignore_properties: (sequence of) `str`, optional}}": """ignore_properties: (sequence of) `str`, optional
                The names of properties to omit from the
                comparison.""",
    # inplace
    "{{inplace: `bool`, optional}}": """inplace: `bool`, optional
                If True then do the operation in-place and return
                `None`.""",
    # ignore_type
    "{{ignore_type: `bool`, optional}}": """ignore_type: `bool`, optional
                Any type of object may be tested but, in general,
                equality is only possible with another `{{class}}`
                instance, or a subclass of one. If *ignore_type* is
                True then ``{{package}}.{{class}}(source=other)`` is
                tested, rather than the ``other`` defined by the
                *other* parameter.""",
    # indent
    "{{indent: `int`, optional}}": """indent: `int`, optional
                Indent each line by this many spaces. By default no
                indentation is applied. Ignored if *string* is
                False.""",
    # name
    "{{name: `str`, optional}}": """name: `str`, optional
                The name of the `{{class}}` instance created by the
                returned commands.

                *Parameter example:*
                  ``name='var1'``""",
    # namespace
    "{{namespace: `str`, optional}}": """namespace: `str`, optional
                The name space containing classes of the {{package}}
                package. This is prefixed to the class name in
                commands that instantiate instances of {{package}}
                objects. By default, or if `None`, the name space is
                assumed to be consistent with {{package}} being
                imported as ``import {{package}}``.

                *Parameter example:*
                  If {{package}} was imported as ``import {{package}}
                  as xyz`` then set ``namespace='xyz'``

                *Parameter example:*
                  If {{package}} was imported as ``from {{package}}
                  import *`` then set ``namespace=''``""",
    # representative_data
    "{{representative_data: `bool`, optional}}": """representative_data: `bool`, optional
                Return one-line representations of `Data` instances,
                which are not executable code but prevent the data
                being converted in its entirety to a string
                representation.""",
    # rtol
    "{{rtol: number, optional}}": """rtol: number, optional
                The tolerance on relative differences between real
                numbers. The default value is set by the
                `{{package}}.rtol` function.""",
    # string
    "{{string: `bool`, optional}}": """string: `bool`, optional
                If False then return each command as an element of a
                `list`. By default the commands are concatenated into
                a string, with a new line inserted between each
                command.""",
    # verbose
    "{{verbose: `int` or `str` or `None`, optional}}": """verbose: `int` or `str` or `None`, optional
                If an integer from ``-1`` to ``3``, or an equivalent
                string equal ignoring case to one of:

                * ``'DISABLE'`` (``0``)
                * ``'WARNING'`` (``1``)
                * ``'INFO'`` (``2``)
                * ``'DETAIL'`` (``3``)
                * ``'DEBUG'`` (``-1``)

                set for the duration of the method call only as the
                minimum cut-off for the verboseness level of displayed
                output (log) messages, regardless of the
                globally-configured `{{package}}.log_level`. Note that
                increasing numerical value corresponds to increasing
                verbosity, with the exception of ``-1`` as a special
                case of maximal and extreme verbosity.

                Otherwise, if `None` (the default value), output
                messages will be shown according to the value of the
                `{{package}}.log_level` setting.

                Overall, the higher a non-negative integer or
                equivalent string that is set (up to a maximum of
                ``3``/``'DETAIL'``) for increasing verbosity, the more
                description that is printed to convey information
                about the operation.""",
    # d
    "{{d: `int`}}": """d: `int`
                The position in the tie points array of the subsampled
                dimension being interpolated.""",
    # d1
    "{{d1: `int`}}": """d1: `int`
                The position of subsampled dimension 1 (in the sense
                of CF Appendix J Figure J.2) in the full tie points
                array.""",
    # d2
    "{{d2: `int`}}": """d2: `int`
                The position of subsampled dimension 2 (in the sense
                of CF Appendix J Figure J.2) in the full tie points
                array.""",
    # s
    "{{s: array_like, optional}}": """s: array_like, optional
                If set to a single number in the range [0, 1] then
                this value is returned for the interpolation
                coefficient ``s``. By default ``s`` is calculated for
                each of the interpolation subarea's uncompressed
                locations.""",
    # s_i
    "{{s_i: array_like}}": """s_i: array_like
                A value for the interpolation coefficient ``s`` for
                the subsampled dimension that corresponds to a
                particular uncompressed location between the two tie
                points.

                *Parameter example:*
                  ``0.5``""",
    # location_use_3d_cartesian
    "{{location_use_3d_cartesian: `numpy.ndarray` or `None`}}": """location_use_3d_cartesian: `numpy.ndarray` or `None`
                The boolean interpolation parameter
                ``location_use_3d_cartesian``, with the same number of
                dimensions in the same order as the full tie points
                array. True values indicate that interpolation is
                carried out in three-dimensional cartesian
                coordinates, as opposed to latitude-longitude
                coordinates. If `None` then an exception is raised.""",
    # construct selection identity
    "{{construct selection identity}}": """A construct has a number of string-valued identities
                defined by its `!identities` method, and is selected
                if any of them match the *identity*
                parameter. *identity* may be a string that equals one
                of a construct's identities; or a `re.Pattern` object
                that matches one of a construct's identities via
                `re.search`.

                Note that in the output of a `dump` method or `print`
                call, a metadata construct is always described by one
                of its identities, and so this description may always
                be used as an *identity* value.""",
    # domain axis selection identity
    "{{domain axis selection identity}}": """A domain axis construct has a number of string-valued
                identities (defined by its `!identities` method) and
                is selected if any of them match the *identity*
                parameter.  *identity* may be a string that equals one
                of a construct's identities; or a `re.Pattern` object
                that matches one of a construct's identities via
                `re.search`.""",
    # returns creation_commands
    "{{returns creation_commands}}": """`str` or `list`
                The commands in a string, with a new line inserted
                between each command. If *string* is False then the
                separate commands are returned as each element of a
                `list`.""",
    # returns dump
    "{{returns dump}}": """`str` or `None`
                The description. If *display* is True then the
                description is printed and `None` is
                returned. Otherwise the description is returned as a
                string.""",
    # generator: `bool`, optional
    "{{generator: `bool`, optional}}": """generator: `bool`, optional
                If True then return a generator for the identities,
                rather than a list. This can give improved performance
                if iteration through the generator can be stopped
                before all identities have been computed.

                .. versionadded:: (cfdm) 1.8.9.0""",
    # filter_kwargs: optional
    "{{filter_kwargs: optional}}": """filter_kwargs: optional
                Keyword arguments as accepted by `Constructs.filter`
                that define additional construct selection
                criteria.""",
    # key: `bool`, optional
    "{{key: `bool`, optional}}": """key: `bool`, optional
                If True then return the selected construct
                identifier. By default the construct itself is
                returned.""",
    # item: `bool`, optional
    "{{item: `bool`, optional}}": """item: `bool`, optional
                If True then return as a tuple the selected construct identifier
                and the construct itself. By default only the construct
                itself is returned. If *key* is True then *item* is
                ignored.""",
    # chunks subarrays
    "{{subarrays chunks: ``-1`` or sequence, optional}}": """chunks: ``-1`` or sequence, optional
                Define the subarray shapes.

                Must be either ``-1``, the default, meaning that all
                non-compressed dimensions in each subarray have the
                maximum possible size; or a sequence of `tuple`. In
                the latter case *chunks* must be allowed as an input
                to `subarray_shapes`.

                A valid *chunks* specification may always be found by
                normalising an output of `subarray_shapes` with
                `dask.array.core.normalize_chunks` as follows:
                ``chunks =
                dask.array.core.normalize_chunks(a.subarray_shapes(...),
                shape=a.shape, dtype=a.dtype)``.""",
    # chunks subarray_shapes
    "{{subarray_shapes chunks: `int`, sequence, `dict`, or `str`, optional}}": """chunks: `int`, sequence, `dict` or `str`, optional
                Define the subarray shapes.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                The subarray sizes implied by *chunks* for a dimension
                that has been compressed are ignored and replaced with
                values that are implied by the decompression algorithm, so
                their specification is arbitrary.

                By default, *chunks* is ``-1``, meaning that all
                non-compressed dimensions in each subarray have the
                maximum possible size.""",
    # clear
    "{{clear: `bool` optional}}": """clear: `bool` optional
                If True then remove any stored original file
                names. This will also clear original file names from
                any ancillary data information, such as a count
                variable required for compressed data.""",
    # define
    "{{define: (sequence of) `str`, optional}}": """define: (sequence of) `str`, optional
                Set these original file names, removing any already
                stored.  The original file names of any constituent
                parts are not set. Can't be used with the *update*
                parameter.""",
    # update
    "{{update: (sequence of) `str`, optional}": """update: (sequence of) `str`, optional
                Add these original file names to those already
                stored. The original file names of any constituent
                parts are not updated. Can't be used with the *define*
                parameter.""",
    # hdf5 chunksizes
    "{{hdf5 chunksizes}}": """chunksizes: `None` or `str` or `int` or `float` or `dict` or a sequence
                Set the chunking strategy for writing to a netCDF4
                file. One of:

                * `None`: No HDF5 chunking strategy has been
                  defined. The chunking strategy will be determined at
                  write time by `{{package}}.write`.

                * ``'contiguous'``: The data will be written to the
                  file contiguously, i.e. no chunking.

                * `int` or `float` or `str`: The size in bytes of the
                  HDF5 chunks. A floating point value is rounded down
                  to the nearest integer, and a string represents a
                  quantity of byte units. "Square-like" chunk shapes
                  are preferred, maximising the amount of chunks that
                  are completely filled with data values (see the
                  `{{package}}.write` *hdf5_chunks* parameter for
                  details). For instance a chunksize of 1024 bytes may
                  be specified with any of ``1024``, ``1024.9``,
                  ``'1024'``, ``'1024.9'``, ``'1024 B'``, ``'1 KiB'``,
                  ``'0.0009765625 MiB'``, etc. Recognised byte units
                  are (case insensitive): ``B``, ``KiB``, ``MiB``,
                  ``GiB``, ``TiB``, ``PiB``, ``KB``, ``MB``, ``GB``,
                  ``TB``, and ``PB``. Spaces in strings are optional.

                * sequence of `int` or `None`: The maximum number of
                  array elements in a chunk along each data axis,
                  provided in the same order as the data axes. Values
                  are automatically limited to the full size of their
                  corresponding data axis, but the special values
                  `None` or ``-1`` may be used to indicate the full
                  axis size. This chunking strategy may get
                  automatically modified by methods that change the
                  data shape (such as `insert_dimension`).

                * `dict`: The maximum number of array elements in a
                  chunk along the axes specified by the dictionary
                  keys. Integer values are automatically limited to
                  the full size of their corresponding data axis, and
                  the special values `None` or ``-1`` may be used to
                  indicate the full axis size. The chunk size for an
                  unspecified axis defaults to an existing chunk size
                  for that axis, if there is one, or else the axis
                  size. This chunking strategy may get automatically
                  modified by methods that change the data shape (such
                  as `insert_dimension`).""",
    # hdf5 todict
    "{{hdf5 todict: `bool`, optional}}": """todict: `bool`, optional
                If True then the HDF5 chunk sizes are returned in a
                `dict` keyed by their axis positions. If False (the
                default) then the HDF5 chunking strategy is returned
                in the same form that it was set (i.e. as `None`,
                `int`, `str`, or `tuple`).""",
    # Returns nc_hdf5_chunksizes
    "{{Returns nc_hdf5_chunksizes}}": """`None` or `str` or `int` or `dict` or `tuple` of `int`
                The current chunking strategy when writing to a
                netCDF4 file. One of:

                * `None`: No HDF5 chunking strategy has been
                  defined. The chunking strategy will be determined at
                  write time by `{{package}}.write`.

                * ``'contiguous'``: The data will be written to the
                  file contiguously, i.e. no chunking.

                * `int` or `str`: The size in bytes of the HDF5
                  chunks. A string represents a quantity of byte
                  units. "Square-like" chunk shapes are preferred,
                  maximising the amount of chunks that are completely
                  filled with data values (see the `{{package}}.write`
                  *hdf5_chunks* parameter for details). For instance a
                  chunksize of 1024 bytes may be specified with any of
                  ``1024``, ``'1024'``, ``'1024 B'``, ``'1 KiB'``,
                  ``'0.0009765625 MiB'``, etc. Recognised byte units
                  are (case insensitive): ``B``, ``KiB``, ``MiB``,
                  ``GiB``, ``TiB``, ``PiB``, ``KB``, ``MB``, ``GB``,
                  ``TB``, and ``PB``.

                * `tuple` of `int`: The maximum number of array
                  elements in a chunk along each data axis. This
                  chunking strategy may get automatically modified by
                  methods that change the data shape (such as
                  `insert_dimension`).

                * `dict`: If *todict* is True, the maximum number of
                  array elements in a chunk along each axis. This
                  chunking strategy may get automatically modified by
                  methods that change the data shape (such as
                  `insert_dimension`).""",
    # init source
    "{{init compressed_dimensions: `dict`}}": """compressed_dimensions: `dict`
                Mapping of compressed to uncompressed dimensions.

                A dictionary key is a position of a dimension in the
                compressed data, with a value of the positions of the
                corresponding dimensions in the uncompressed
                data. Compressed array dimensions that are not
                compressed must be omitted from the mapping.""",
    # init start_index
    "{{init start_index: `int`}}": """start_index: `int`
                The base of the indices provided by the integer index
                array. Must be ``0`` or ``1`` for zero- or one-based
                indices respectively.""",
    # init cell_dimension
    "{{init cell_dimension: `int`}}": """cell_dimension: `int`
                The position of the *data* dimension that indexes the
                cells, either ``0`` or ``1``.""",
    # init mask
    "{{init mask: `bool`, optional}}": """mask: `bool`, optional
                If True (the default) then mask by convention when
                reading data from disk.""",
    # init unpack
    "{{init unpack: `bool`, optional}}": """unpack: `bool`, optional
                If True (the default) then unpack by convention when
                reading data from disk.""",
    # init attributes
    "{{init attributes: `dict` or `None`, optional}}": """attributes: `dict` or `None`, optional
                Provide netCDF attributes for the data as a dictionary
                of key/value pairs.""",
    # init storage_options
    "{{init storage_options: `dict` or `None`, optional}}": """storage_options: `dict` or `None`, optional
                Key/value pairs to be passed on to the creation of
                `s3fs.S3FileSystem` file systems to control the
                opening of files in S3 object stores. Ignored for
                files not in an S3 object store, i.e. those whose
                names do not start with ``s3:``.

                By default, or if `None`, then *storage_options* is
                taken as ``{}``.

                If the ``'endpoint_url'`` key is not in
                *storage_options* or is not in a dictionary defined by
                the ``'client_kwargs`` key (which is always the case
                when *storage_options* is `None`), then one will be
                automatically inserted for accessing an S3 file. For
                example, for a file name of
                ``'s3://store/data/file.nc'``, an ``'endpoint_url'``
                key with value ``'https://store'`` would be created.

                *Parameter example:*
                  For a file name of ``'s3://store/data/file.nc'``,
                  the following are equivalent: ``None``, ``{}``, and
                  ``{'endpoint_url': 'https://store'}``,
                  ``{'client_kwargs': {'endpoint_url':
                  'https://store'}}``

                *Parameter example:*
                  ``{'key': 'scaleway-api-key...', 'secret':
                  'scaleway-secretkey...', 'endpoint_url':
                  'https://s3.fr-par.scw.cloud', 'client_kwargs':
                  {'region_name': 'fr-par'}}``""",
    # _force_mask_hardness
    "{{_force_mask_hardness: `bool`, optional}}": """_force_mask_hardness: `bool`, optional
                If True (the default) then force the mask hardness of
                the returned Dask graph to be that given by the
                `hardmask` attribute. If False then the mask hardness
                may or may not be correct, depending on the nature of
                the stack of previously defined lazy operations.

                Set to False if the intention is to just inspect the
                state of the Dask graph, or is to add to the returned
                Dask graph further operations to which can correctly
                manage the mask hardness.""",
    # _force_to_memory
    "{{_force_to_memory: `bool`, optional}}": """_force_to_memory: `bool`, optional
                If True (the default) then force the data resulting
                from computing the returned Dask graph to be in
                memory. If False then the data resulting from
                computing the Dask graph may or may not be in memory,
                depending on the nature of the stack of previously
                defined lazy operations.

                Set to False if the intention is to just inspect the
                state of the Dask graph, or is to add to the returned
                Dask graph further operations to which can handle any
                required conversion to data in memory.""",
    # chunks
    "{{chunks: `int`, `tuple`, `dict` or `str`, optional}}": """chunks: `int`, `tuple`, `dict` or `str`, optional
                Specify the chunking of the underlying dask array.

                Any value accepted by the *chunks* parameter of the
                `dask.array.from_array` function is allowed.

                By default, ``"auto"`` is used to specify the array
                chunking, which uses a chunk size in bytes defined by
                the `{{package}}.chunksize` function, preferring square-like
                chunk shapes.

                *Parameter example:*
                  A blocksize like ``1000``.

                *Parameter example:*
                  A blockshape like ``(1000, 1000)``.

                *Parameter example:*
                  Explicit sizes of all blocks along all dimensions
                  like ``((1000, 1000, 500), (400, 400))``.

                *Parameter example:*
                  A size in bytes, like ``"100MiB"`` which will choose
                  a uniform block-like shape, preferring square-like
                  chunk shapes.

                *Parameter example:*
                  A blocksize of ``-1`` or `None` in a tuple or
                  dictionary indicates the size of the corresponding
                  dimension.

                *Parameter example:*
                  Blocksizes of some or all dimensions mapped to
                  dimension positions, like ``{1: 200}``, or ``{0: -1,
                  1: (400, 400)}``.""",
    # threshold
    "{{threshold: `int`, optional}}": """threshold: `int`, optional
                The graph growth factor under which we don't bother
                introducing an intermediate step. See
                `dask.array.rechunk` for details.""",
    # block_size_limit
    "{{block_size_limit: `int`, optional}}": """block_size_limit: `int`, optional
                The maximum block size (in bytes) we want to produce,
                as defined by the `{{package}}.chunksize` function.""",
    # balance
    "{{balance: `bool`, optional}}": """balance: `bool`, optional
                If True, try to make each chunk the same size. By
                default this is not attempted.

                This means ``balance=True`` will remove any small
                leftover chunks, so using ``d.rechunk(chunks=len(d) //
                N, balance=True)`` will almost certainly result in
                ``N`` chunks.""",
    # collapse keepdims
    "{{collapse keepdims: `bool`, optional}}": """keepdims: `bool`, optional
                By default, the axes which are collapsed are left in
                the result as dimensions with size one, so that the
                result will broadcast correctly against the input
                array. If set to False then collapsed axes are removed
                from the data.""",
    # collapse axes
    "{{collapse axes: (sequence of) `int`, optional}}": """axes: (sequence of) `int`, optional
                The axes to be collapsed. By default all axes are
                collapsed, resulting in output with size 1. Each axis
                is identified by its positive or negative integer
                position. If *axes* is an empty sequence then the
                collapse is applied to each scalar element and the
                result has the same shape as the input data.""",
    # collapse squeeze
    "{{collapse squeeze: `bool`, optional}}": """squeeze: `bool`, optional
                By default, the axes which are collapsed are left in
                the result as dimensions with size one, so that the
                result will broadcast correctly against the input
                array. If set to True then collapsed axes are removed
                from the data.""",
    # split_every
    "{{split_every: `int` or `dict`, optional}}": """split_every: `int` or `dict`, optional
                Determines the depth of the `dask` recursive
                aggregation. If set to or more than the number of
                input Dask chunks, the aggregation will be performed
                in two steps, one partial collapse per input chunk and
                a single aggregation at the end. If set to less than
                that, an intermediate aggregation step will be used,
                so that any of the intermediate or final aggregation
                steps operates on no more than ``split_every``
                inputs. The depth of the aggregation graph will be
                :math:`log_{split\_every}(\textnormal{input chunks
                along reduced axes})`. Setting to a low value can
                reduce cache size and network transfers, at the cost
                of more CPU and a larger dask graph.

                By default, `dask` heuristically decides on a good
                value. A default can also be set globally with the
                ``split_every`` key in `dask.config`. See
                `dask.array.reduction` for details.""",
    # _get_array index
    "{{index: `tuple` or `None`, optional}}": """index: `tuple` or `None`, optional
               Provide the indices that define the subspace. If `None`
               then the `index` attribute is used.""",
    # pad_width
    "{{pad_width: sequence of `int`, optional}}": """pad_width: sequence of `int`, optional
                Number of values to pad before and after the edges of
                the axis.""",
    # to_size
    "{{to_size: `int`, optional}}": """to_size: `int`, optional
                Pad the axis after so that the new axis has the given
                size.""",
    # cull_graph
    "{{cull_graph: `bool`, optional}}": """cull_graph: `bool`, optional
                If True then unnecessary tasks are removed (culled)
                from each array's dask graph before
                concatenation. This process can have a considerable
                overhead but can sometimes improve the overall
                performance of a workflow. If False (the default) then
                dask graphs are not culled. See
                `dask.optimization.cull` for details.""",
    # relaxed_units
    "{{relaxed_units: `bool`, optional}}": """relaxed_units: `bool`, optional
                If True then allow the concatenation of data with
                invalid but otherwise equal units. By default, if any
                data array has invalid units then the concatenation
                will fail. A `Units` object is considered to be
                invalid if its `!isvalid` attribute is `False`.""",
    # relaxed_units
    "{{concatenate copy: `bool`, optional}}": """copy: `bool`, optional
                If True (the default) then make copies of the
                `{{class}}` objects prior to the concatenation,
                thereby ensuring that the input constructs are not
                changed by the concatenation process. If False then
                some or all input constructs might be changed
                in-place, but the concatenation process will be
                faster.""",
    # normalise
    "{{normalise: `bool`, optional}}": """normalise: `bool`, optional
                If True then normalise to an absolute path. If False
                (the default) then no normalisation is done.""",
    # replace old
    "{{replace old: `str` or `None`, optional}}": """old: `str` or `None`, optional
                The base directory structure to be replaced by
                *new*. If `None` (the default) or an empty string, and
                *normalise* is False, then *new* (if set) is prepended
                to each file name.""",
    # replace new
    "{{replace new: `str` or `None`, optional}}": """new: `str` or `None`, optional
                The new directory that replaces the base directory
                structure identified by *old*. If `None` (the default)
                or an empty string, then *old* (if set) is replaced
                with an empty string.""",
    # replace normalise
    "{{replace normalise: `bool`, optional}}": """normalise: `bool`, optional
                If True then *old* and *new* directories, and the file
                names, are normalised to absolute paths prior to the
                replacement. If False (the default) then no
                normalisation is done.""",
    # ----------------------------------------------------------------
    # Method description substitutions (4 levels of indentation)
    # ----------------------------------------------------------------
    # Returns constructs
    "{{Returns constructs}}": """The selected constructs in a new `Constructs` object,
                unless modified by any *filter_kwargs* parameters. The
                returned object will contain no constructs if none
                were selected.""",
    # Returns construct
    "{{Returns construct}}": """The selected construct, or its identifier if *key* is
                True, or a tuple of both if *item* is True.""",
    # string value match
    "{{value match}}": """A value may be any object that can match via the
                ``==`` operator, or a `re.Pattern` object that matches
                via its `~re.Pattern.search` method.""",
    # displayed identity
    "{{displayed identity}}": """Note that in the output of a `dump` method or `print`
                call, a construct is always described by an identity
                that will select it.""",
    # auto:
    "{{shapes auto}}": """If ``'auto'`` then the shapes along non-compressed
                dimensions are not created, those dimensions' elements
                being replaced with ``'auto'`` instead.""",
    # Returns original filenames
    "{{Returns original filenames}}": """The original file names in normalised absolute
                form. If there are no original files then an empty
                `set` will be returned.""",
}



import datetime
import faulthandler
import os
import unittest

faulthandler.enable()  # to debug seg faults and timeouts

import cfdm


class FieldAncillaryTest(unittest.TestCase):
    """Unit test for the FieldAncillary class."""

    filename = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "test_file.nc"
    )

    #    f = cfdm.read(filename)[0]

    def setUp(self):
        """Preparations called immediately before each test method."""
        # Disable log messages to silence expected warnings
        cfdm.log_level("DISABLE")
        # Note: to enable all messages for given methods, lines or
        # calls (those without a 'verbose' option to do the same)
        # e.g. to debug them, wrap them (for methods, start-to-end
        # internally) as follows:
        #
        # cfdm.LOG_LEVEL('DEBUG')
        # < ... test code ... >
        # cfdm.log_level('DISABLE')
        self.f = cfdm.read(self.filename)[0]

    def test_FieldAncillary__repr__str__dump(self):
        """Test all means of FieldAncillary inspection."""
        f = self.f.copy()
        x = f.field_ancillaries("ancillaryA").value()

        _ = repr(x)
        _ = str(x)
        self.assertIsInstance(x.dump(display=False), str)

    def test_FieldAncillary_source(self):
        """Test the source keyword argument to FieldAncillary."""
        f = self.f.copy()

        a = f.auxiliary_coordinates("latitude").value()
        cfdm.FieldAncillary(source=a)

    def test_FieldAncillary_properties(self):
        """Test the property access methods of FieldAncillary."""
        f = self.f.copy()
        x = f.domain_ancillaries("ncvar%a").value()
        x = cfdm.FieldAncillary(source=x)

        x.set_property("long_name", "qwerty")

        self.assertEqual(x.get_property("long_name"), "qwerty")
        self.assertEqual(x.del_property("long_name"), "qwerty")
        self.assertIsNone(x.get_property("long_name", None))
        self.assertIsNone(x.del_property("long_name", None))

    def test_FieldAncillary_insert_dimension(self):
        """Test the `insert_dimension` FieldAncillary method."""
        f = self.f.copy()
        d = f.dimension_coordinates("grid_longitude").value()
        x = cfdm.FieldAncillary(source=d)

        self.assertEqual(x.shape, (9,))

        y = x.insert_dimension(0)
        self.assertEqual(y.shape, (1, 9))

        x.insert_dimension(-1, inplace=True)
        self.assertEqual(x.shape, (9, 1))

    def test_FieldAncillary_transpose(self):
        """Test the transpose FieldAncillary method."""
        f = self.f.copy()
        a = f.auxiliary_coordinates("longitude").value()
        x = cfdm.FieldAncillary(source=a)

        self.assertEqual(x.shape, (9, 10))

        y = x.transpose()
        self.assertEqual(y.shape, (10, 9))

        x.transpose([1, 0], inplace=True)
        self.assertEqual(x.shape, (10, 9))

    def test_FieldAncillary_squeeze(self):
        """Test the squeeze FieldAncillary method."""
        f = self.f.copy()
        a = f.auxiliary_coordinates("longitude").value()
        x = cfdm.FieldAncillary(source=a)

        x.insert_dimension(1, inplace=True)
        x.insert_dimension(0, inplace=True)

        self.assertEqual(x.shape, (1, 9, 1, 10))

        y = x.squeeze()
        self.assertEqual(y.shape, (9, 10))

        x.squeeze(2, inplace=True)
        self.assertEqual(x.shape, (1, 9, 10))


if __name__ == "__main__":
    print("Run date:", datetime.datetime.now())
    cfdm.environment()
    print()
    unittest.main(verbosity=2)



"""General functions useful for `Data` functionality."""

from functools import lru_cache, partial
from itertools import product

import cftime
import dask.array as da
import numpy as np
from dask.core import flatten

from ..units import Units

_default_calendar = "standard"


def allclose(x, y, masked_equal=True, rtol=None, atol=None):
    """An effective dask.array.ma.allclose method.

    True if two dask arrays are element-wise equal within a tolerance.

    Equivalent to allclose except that masked values are treated as
    equal (default) or unequal, depending on the masked_equal
    argument.

    Define an effective da.ma.allclose method here because one is
    currently missing in the Dask codebase.

    Note that all default arguments are the same as those provided to
    the corresponding NumPy method (see the `numpy.ma.allclose` API
    reference).

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        x: a dask array to compare with y

        y: a dask array to compare with x

        masked_equal: `bool`, optional
            Whether masked values in a and b are considered equal
            (True) or not (False). They are considered equal by
            default.

        {{rtol: number, optional}}

        {{atol: number, optional}}

    :Returns:

        `bool`
            A Boolean value indicating whether or not the two dask
            arrays are element-wise equal to the given *rtol* and
            *atol* tolerance.

    """
    if rtol is None or atol is None:
        raise ValueError(
            "Must provide numeric values for the rtol and atol keywords"
        )

    # Must pass rtol=rtol, atol=atol in as kwargs to allclose, rather than it
    # using those in local scope from the outer function arguments, because
    # Dask's internal algorithms require these to be set as parameters.
    def allclose(a_blocks, b_blocks, rtol=rtol, atol=atol):
        """Run `ma.allclose` across multiple blocks over two arrays."""
        result = True
        # Handle scalars, including 0-d arrays, for which a_blocks and
        # b_blocks will have the corresponding type and hence not be iterable.
        # With this approach, we avoid inspecting sizes or lengths, and for
        # the 0-d array blocks the following iteration can be used unchanged
        # and will only execute once with block sizes as desired of:
        # (np.array(<int size>),)[0] = array(<int size>). Note
        # can't check against more general case of collections.abc.Iterable
        # because a 0-d array is also iterable, but in practice always a list.
        if not isinstance(a_blocks, list):
            a_blocks = (a_blocks,)
        if not isinstance(b_blocks, list):
            b_blocks = (b_blocks,)

        # Note: If a_blocks or b_blocks has more than one chunk in
        #       more than one dimension they will comprise a nested
        #       sequence of sequences, that needs to be flattened so
        #       that we can safely iterate through the actual numpy
        #       array elements.

        for a, b in zip(flatten(a_blocks), flatten(b_blocks)):
            result &= np.ma.allclose(
                a, b, masked_equal=masked_equal, rtol=rtol, atol=atol
            )

        return result

    axes = tuple(range(x.ndim))
    return da.blockwise(
        allclose, "", x, axes, y, axes, dtype=bool, rtol=rtol, atol=atol
    )


def collapse(
    func,
    d,
    axis=None,
    keepdims=True,
    split_every=None,
):
    """Collapse data in-place using a given funcion.

     .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        func: callable
            The function that collapses the underlying `dask` array of
            *d*. Must have the minimum signature (parameters and
            default values) ``func(dx, axis=None, keepdims=False,
            mtol=None, split_every=None)``, where ``dx`` is a the dask
            array contained in *d*.

        d: `Data`
            The data to be collapsed.

        axis: (sequence of) int, optional
            The axes to be collapsed. By default all axes are
            collapsed, resulting in output with size 1. Each axis is
            identified by its integer position. If *axes* is an empty
            sequence then the collapse is applied to each scalar
            element and the reuslt has the same shape as the input
            data.

        keepdims: `bool`, optional
            By default, the axes which are collapsed are left in the
            result as dimensions with size one, so that the result
            will broadcast correctly against the input array. If set
            to False then collapsed axes are removed from the data.

        split_every: `int` or `dict`, optional
            Determines the depth of the recursive aggregation. See
            `dask.array.reduction` for details.

    :Returns:

        `Data`
            The collapsed data.

    """
    dx = d.to_dask_array()
    original_shape = dx.shape

    # Parse the axes. By default flattened input is used.
    if axis is None:
        iaxes = tuple(range(dx.ndim))
    else:
        try:
            iaxes = d._parse_axes(axis)
        except ValueError as error:
            raise ValueError(f"Can't min data: {error}")

    dx = func(dx, axis=iaxes, keepdims=keepdims, split_every=split_every)
    d._set_dask(dx)

    # Remove collapsed axis identifiers
    if iaxes and not keepdims:
        d._axes = [axis for i, axis in enumerate(d._axes) if i not in iaxes]

    # Update the HDF5 chunking strategy
    chunksizes = d.nc_hdf5_chunksizes()
    if (
        chunksizes
        and isinstance(chunksizes, tuple)
        and dx.shape != original_shape
    ):
        if not keepdims:
            chunksizes = [
                size for i, size in enumerate(chunksizes) if i not in iaxes
            ]

        d.nc_set_hdf5_chunksizes(chunksizes)

    return d


def is_numeric_dtype(array):
    """True if the given array is of a numeric or boolean data type.

    .. versionadded:: (cfdm) 1.11.2.0

        :Parameters:

            array: numpy-like array

        :Returns:

            `bool`
                Whether or not the array holds numeric elements.

    **Examples**

    >>> a = np.array([0, 1, 2])
    >>> cfdm.data.utils.is_numeric_dtype(a)
    True
    >>> a = np.array([False, True, True])
    >>> cfdm.data.utils.is_numeric_dtype(a)
    True
    >>> a = np.array(["a", "b", "c"], dtype="S1")
    >>> cfdm.data.utils.is_numeric_dtype(a)
    False
    >>> a = np.ma.array([10.0, 2.0, 3.0], mask=[1, 0, 0])
    >>> cfdm.data.utils.is_numeric_dtype(a)
    True
    >>> a = np.array(10)
    >>> cfdm.data.utils.is_numeric_dtype(a)
    True
    >>> a = np.empty(1, dtype=object)
    >>> cfdm.data.utils.is_numeric_dtype(a)
    False

    """
    dtype = array.dtype

    # This checks if the dtype is either a standard "numeric" type
    # (i.e. int types, floating point types or complex floating point
    # types) or Boolean, which are effectively a restricted int type
    # (0 or 1). We determine the former by seeing if it sits under the
    # 'np.number' top-level dtype in the NumPy dtype hierarchy; see
    # the 'Hierarchy of type objects' figure diagram under:
    # https://numpy.org/doc/stable/reference/arrays.scalars.html#scalars
    return np.issubdtype(dtype, np.number) or np.issubdtype(dtype, np.bool_)


def convert_to_datetime(a, units):
    """Convert a dask array of numbers to one of date-time objects.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso `convert_to_reftime`

    :Parameters:

        a: `dask.array.Array`
            The input numeric reference time values.

        units: `Units`
            The reference time units that define the output
            date-time objects.

    :Returns:

        `dask.array.Array`
            A new dask array containing date-time objects.

    **Examples**

    >>> import dask.array as da
    >>> d = da.from_array(2.5)
    >>> e = cfdm.data.utils.convert_to_datetime(d, cfdm.Units("days since 2000-12-01"))
    >>> print(e.compute())
    2000-12-03 12:00:00

    """
    return a.map_blocks(
        partial(rt2dt, units_in=units),
        dtype=object,
        meta=np.array((), dtype=object),
    )


def convert_to_reftime(a, units=None, first_value=None):
    """Convert date-times to floating point reference times.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso `convert_to_datetime`

    :Parameters:

        a: `dask.array.Array`
            An array of string or object date-times

        units: `Units`, optional
            Specify the units for the output reference time values. By
            default the units are inferred from the first non-missing
            value in the array, or set to ``<Units: days since
            1970-01-01 gregorian>`` if all values are missing.

        first_value: optional
            If set, then assumed to be equal to the first non-missing
            value of the array, thereby removing the need to find it
            by inspection of *a*, which may be expensive. By default
            the first non-missing value is found from *a*.

    :Returns:

        (`dask.array.Array`, `Units`)
            The reference times, and their units.

    >>> import dask.array as da
    >>> d = da.from_array(2.5)
    >>> e = cfdm.data.utils.convert_to_datetime(d, Units("days since 2000-12-01"))

    >>> f, u = cfdm.data.utils.convert_to_reftime(e)
    >>> f.compute()
    0.5
    >>> u
    <Units: days since 2000-12-3 standard>

    >>> f, u = cfdm.data.utils.convert_to_reftime(e, Units("days since 1999-12-01"))
    >>> f.compute()
    368.5
    >>> u
    <Units: days since 1999-12-01 standard>

    """
    kind = a.dtype.kind
    if kind in "US":
        # Convert date-time strings to reference time floats
        if not units:
            first_value = first_non_missing_value(a, cached=first_value)
            if first_value is not None:
                YMD = str(first_value).partition("T")[0]
            else:
                YMD = "1970-01-01"

            units = Units(
                "days since " + YMD,
                getattr(units, "calendar", _default_calendar),
            )

        a = a.map_blocks(
            partial(st2rt, units_in=units, units_out=units), dtype=float
        )

    elif kind == "O":
        # Convert date-time objects to reference time floats
        first_value = first_non_missing_value(a, cached=first_value)
        if first_value is not None:
            x = first_value
        else:
            x = cftime.DatetimeGregorian(1970, 1, 1)

        x_since = "days since " + "-".join(map(str, (x.year, x.month, x.day)))
        x_calendar = getattr(x, "calendar", "standard")

        d_calendar = getattr(units, "calendar", None)
        d_units = getattr(units, "units", None)

        if x_calendar != "":
            if not units:
                d_calendar = x_calendar
            elif not units.equivalent(Units(x_since, x_calendar)):
                raise ValueError(
                    "Incompatible units: "
                    f"{units!r}, {Units(x_since, x_calendar)!r}"
                )

        if not units:
            # Set the units to something that is (hopefully) close to
            # all of the datetimes, in an attempt to reduce errors
            # arising from the conversion to reference times
            units = Units(x_since, calendar=d_calendar)
        else:
            units = Units(d_units, calendar=d_calendar)

        # Convert the date-time objects to reference times
        a = a.map_blocks(dt2rt, units_out=units, dtype=float)

    if not units.isreftime:
        raise ValueError(
            f"Can't create a reference time array with units {units!r}"
        )

    return a, units


def first_non_missing_value(a, cached=None, method="index"):
    """Return the first non-missing value of a dask array.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        a: `dask.array.Array`
            The array to be inspected.

        cached: scalar, optional
            If set to a value other than `None`, then return without
            inspecting the array. This allows a previously found first
            value to be used instead of a potentially costly array
            access.

        method: `str`, optional
            Select the method used to find the first non-missing
            value.

            The default ``'index'`` method evaulates sequentially the
            elements of the flattened array and returns when the first
            non-missing value is found.

            The ``'mask'`` method finds the first non-missing value of
            the flattened array as that which has the same location as
            the first `False` element of the flattened array mask.

            It is considered likely that the ``'index'`` method is
            fastest for data for which the first element is not
            missing, but this may not always be the case.

    :Returns:

            If set, then *cached* is returned. Otherwise returns the
            first non-missing value of *a*, or `None` if there isn't
            one.

    **Examples**

    >>> import dask.array as da
    >>> d = da.arange(8).reshape(2, 4)
    >>> print(d.compute())
    [[0 1 2 3]
     [4 5 6 7]]
    >>> cfdm.data.utils.first_non_missing_value(d)
    0
    >>> cfdm.data.utils.first_non_missing_value(d, cached=99)
    99
    >>> d[0, 0] = cfdm.masked
    >>> cfdm.data.utils.first_non_missing_value(d)
    1
    >>> d[0, :] = cfdm.masked
    >>> cfdm.data.utils.first_non_missing_value(d)
    4
    >>> cfdm.data.utils.first_non_missing_value(d, cached=99)
    99
    >>> d[...] = cfdm.masked
    >>> print(cfdm.data.utils.first_non_missing_value(d))
    None
    >>> print(cfdm.data.utils.first_non_missing_value(d, cached=99))
    99

    """
    if cached is not None:
        return cached

    if method == "index":
        shape = a.shape
        for i in range(a.size):
            index = np.unravel_index(i, shape)
            x = a[index].compute()
            if not (x is np.ma.masked or np.ma.getmask(x)):
                try:
                    return x.item()
                except AttributeError:
                    return x

        return

    if method == "mask":
        mask = da.ma.getmaskarray(a)
        if not a.ndim:
            # Scalar data
            if mask:
                return

            a = a.compute()
            try:
                return a.item()
            except AttributeError:
                return a

        x = a[da.unravel_index(mask.argmin(), a.shape)].compute()
        if x is np.ma.masked:
            return

        try:
            return x.item()
        except AttributeError:
            return x

    raise ValueError(f"Unknown value of 'method': {method!r}")


@lru_cache(maxsize=32)
def generate_axis_identifiers(n):
    """Return new axis identifiers for a given number of axes.

    The names are arbitrary and have no semantic meaning.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        n: `int`
            Generate this number of axis identifiers.

    :Returns:

        `list`
            The new axis identifiers.

    **Examples**

    >>> cfdm.data.creation.generate_axis_identifiers(0)
    []
    >>> cfdm.data.creation.generate_axis_identifiers(1)
    ['dim0']
    >>> cfdm.data.creation.generate_axis_identifiers(3)
    ['dim0', 'dim1', 'dim2']

    """
    return [f"dim{i}" for i in range(n)]


@lru_cache(maxsize=32)
def new_axis_identifier(existing_axes=(), basename="dim"):
    """Return a new, unique axis identifier.

    The name is arbitrary and has no semantic meaning.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        existing_axes: sequence of `str`, optional
            Any existing axis names that are not to be duplicated.

        basename: `str`, optional
            The root of the new axis identifier. The new axis
            identifier will be this root followed by an integer.

    :Returns:

        `str`
            The new axis idenfifier.

    **Examples**

    >>> cfdm.data.utils.new_axis_identifier()
    'dim0'
    >>> cfdm.data.utils.new_axis_identifier(['dim0'])
    'dim1'
    >>> cfdm.data.utils.new_axis_identifier(['dim3'])
    'dim1'
     >>> cfdm.data.utils.new_axis_identifier(['dim1'])
    'dim2'
    >>> cfdm.data.utils.new_axis_identifier(['dim1', 'dim0'])
    'dim2'
    >>> cfdm.data.utils.new_axis_identifier(['dim3', 'dim4'])
    'dim2'
    >>> cfdm.data.utils.new_axis_identifier(['dim2', 'dim0'])
    'dim3'
    >>> cfdm.data.utils.new_axis_identifier(['dim3', 'dim4', 'dim0'])
    'dim5'
    >>> cfdm.data.utils.new_axis_identifier(basename='axis')
    'axis0'
    >>> cfdm.data.utils.new_axis_identifier(basename='axis')
    'axis0'
    >>> cfdm.data.utils.new_axis_identifier(['dim0'], basename='axis')
    'axis1'
    >>> cfdm.data.utils.new_axis_identifier(['dim0', 'dim1'], basename='axis')
    'axis2'

    """
    n = len(existing_axes)
    axis = f"{basename}{n}"
    while axis in existing_axes:
        n += 1
        axis = f"{basename}{n}"

    return axis


def chunk_indices(chunks):
    """Return indices that define each dask chunk.

    .. versionadded:: (cfdm) 1.12.0.0

    .. seealso:: `chunks`

    :Parameters:

        chunks: `tuple`
            The chunk sizes along each dimension, as output by
            `dask.array.Array.chunks`.

    :Returns:

        `itertools.product`
            An iterator over tuples of indices of the data array.

    **Examples**

    >>> chunks = ((1, 2), (9,), (4, 5, 6)))
    >>> for index in cfdm.data.utils.chunk_indices():
    ...     print(index)
    ...
    (slice(0, 1, None), slice(0, 9, None), slice(0, 4, None))
    (slice(0, 1, None), slice(0, 9, None), slice(4, 9, None))
    (slice(0, 1, None), slice(0, 9, None), slice(9, 15, None))
    (slice(1, 3, None), slice(0, 9, None), slice(0, 4, None))
    (slice(1, 3, None), slice(0, 9, None), slice(4, 9, None))
    (slice(1, 3, None), slice(0, 9, None), slice(9, 15, None))

    """
    from dask.utils import cached_cumsum

    cumdims = [cached_cumsum(bds, initial_zero=True) for bds in chunks]
    indices = [
        [slice(s, s + dim) for s, dim in zip(starts, shapes)]
        for starts, shapes in zip(cumdims, chunks)
    ]
    return product(*indices)


def chunk_positions(chunks):
    """Find the position of each chunk.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso:: `chunk_indices`, `chunk_locations`, `chunk_shapes`

    :Parameters:

        chunks: `tuple`
            The chunk sizes along each dimension, as output by
            `dask.array.Array.chunks`.

    **Examples**

    >>> chunks = ((1, 2), (9,), (4, 5, 6))
    >>> for position in cfdm.data.utils.chunk_positions(chunks):
    ...     print(position)
    ...
    (0, 0, 0)
    (0, 0, 1)
    (0, 0, 2)
    (1, 0, 0)
    (1, 0, 1)
    (1, 0, 2)

    """
    return product(*(range(len(bds)) for bds in chunks))


def chunk_shapes(chunks):
    """Find the shape of each chunk.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso:: `chunk_indices`, `chunk_locations`, `chunk_positions`

    :Parameters:

        chunks: `tuple`
            The chunk sizes along each dimension, as output by
            `dask.array.Array.chunks`.

    **Examples**

    >>> chunks = ((1, 2), (9,), (4, 5, 6))
    >>> for shape in cfdm.data.utils.chunk_shapes(chunks):
    ...     print(shape)
    ...
    (1, 9, 4)
    (1, 9, 5)
    (1, 9, 6)
    (2, 9, 4)
    (2, 9, 5)
    (2, 9, 6)

    """
    return product(*chunks)


def chunk_locations(chunks):
    """Find the shape of each chunk.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso:: `chunk_indices`, `chunk_positions`, `chunk_shapes`

    :Parameters:

        chunks: `tuple`
            The chunk sizes along each dimension, as output by
            `dask.array.Array.chunks`.

    **Examples**

    >>> chunks = ((1, 2), (9,), (4, 5, 6))
    >>> for location in cfdm.data.utils.chunk_locations(chunks):
    ...     print(location)
    ...
    ((0, 1), (0, 9), (0, 4))
    ((0, 1), (0, 9), (4, 9))
    ((0, 1), (0, 9), (9, 15))
    ((1, 3), (0, 9), (0, 4))
    ((1, 3), (0, 9), (4, 9))
    ((1, 3), (0, 9), (9, 15))

    """
    from dask.utils import cached_cumsum

    cumdims = [cached_cumsum(bds, initial_zero=True) for bds in chunks]
    locations = [
        [(s, s + dim) for s, dim in zip(starts, shapes)]
        for starts, shapes in zip(cumdims, chunks)
    ]
    return product(*locations)


def normalize_chunks(chunks, shape=None, dtype=None):
    """Normalize chunks to tuple of tuples.

    The shape may contain sizes of ``nan``. This could occur when the
    underlying data is compressed in a way which makes the shape
    impossible to infer without actually uncompressing the data.

    If *shape* contains no ``nan`` sizes then this function is
    identical to `dask.array.core.normalize_chunks`. If it does, then
    the output chunks for each such axis will be ``(nan,)``.

    .. versionadded (cfdm) 1.11.2.0

    :Parameters:

        chunks: tuple, int, dict, or string
            The chunks to be normalized. See
            `dask.array.core.normalize_chunks` for details.

        shape: `tuple`
            The shape of the data.

        dtype: data-type
            The data-type for the data.

    :Returns:

        `tuple`
            The normalized chunks.

    """
    from math import isnan, nan

    from dask.array.core import normalize_chunks

    if not any(map(isnan, shape)):
        return normalize_chunks(chunks, shape=shape, dtype=dtype)

    out = [
        (
            (nan,)
            if isnan(size)
            else normalize_chunks(chunk, shape=(size,), dtype=dtype)[0]
        )
        for chunk, size in zip(chunks, shape)
    ]
    return tuple(out)


def dt2rt(array, units_out):
    """Return numeric time values from datetime objects.

    .. versionadded:: (cfdm) 1.11.2.0

    .. seealso:: `rt2dt`

    :Parameters:

        array: numpy array-like of date-time objects
            The datetime objects must be in UTC with no time-zone
            offset.

        units_out: `Units`
            The units of the numeric time values. If there is a
            time-zone offset in *units_out*, it will be applied to the
            returned numeric values.

    :Returns:

        `numpy.ndarray`
            An array of numbers with the same shape as *array*.

    **Examples**

    >>> print(
    ...   cfdm.data.utils.dt2rt(
    ...     np.ma.array([0, cftime.DatetimeGregorian(2001, 11, 16, 12)], mask=[True, False]),
    ...     units_out=cfdm.Units('days since 2000-01-01')
    ...   )
    ... )
    [-- 685.5]

    """
    isscalar = not np.ndim(array)

    array = cftime.date2num(
        array, units=units_out.units, calendar=units_out._utime.calendar
    )

    if isscalar:
        if array is np.ma.masked:
            array = np.ma.masked_all(())
        else:
            array = np.asanyarray(array)

    return array


def rt2dt(array, units_in):
    """Convert reference times to date-time objects.

    .. versionadded:: (cfdm) 1.11.2.0

    The returned array is always independent.

    .. seealso:: `dt2rt`

    :Parameters:

        array: numpy array-like

        units_in: `Units`

    :Returns:

        `numpy.ndarray`
            An array of `cftime.datetime` objects with the same shape
            as *array*.

    **Examples**

    >>> print(
    ...   cfdm.data.utils.rt2dt(
    ...     np.ma.array([0, 685.5], mask=[True, False]),
    ...     units_in=cfdm.Units('days since 2000-01-01')
    ...   )
    ... )
    [--
     cftime.DatetimeGregorian(2001, 11, 16, 12, 0, 0, 0, has_year_zero=False)]

    """
    if not np.ndim(array) and np.ma.is_masked(array):
        # num2date has issues with scalar masked arrays with a True
        # mask
        return np.ma.masked_all((), dtype=object)

    units = units_in.units
    calendar = getattr(units_in, "calendar", "standard")

    if np.ma.isMA(array):
        # Note: We're going to apply `cftime.num2date` to a non-masked
        #       array and reset the mask afterwards, because numpy
        #       currently (numpy==2.2.3) has a bug that produces a
        #       RuntimeWarning: "numpy/ma/core.py:502: RuntimeWarning:
        #       invalid value encountered in cast fill_value =
        #       np.asarray(fill_value, dtype=ndtype)". See
        #       https://github.com/numpy/numpy/issues/28255 for more
        #       details.
        mask = array.mask
        array = np.array(array)
    else:
        mask = None

    array = cftime.num2date(
        array, units, calendar, only_use_cftime_datetimes=True
    )

    if not isinstance(array, np.ndarray):
        array = np.array(array, dtype=object)

    if mask is not None:
        array = np.ma.array(array, mask=mask)

    return array


def st2datetime(date_string, calendar=None):
    """Parse an ISO 8601 date-time string into a `cftime` object.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        date_string: `str`

    :Returns:

        `cftime.datetime`

    """
    if date_string.count("-") != 2:
        raise ValueError(
            "Input date-time string must contain at least a year, a month "
            "and a day"
        )

    x = cftime._parse_date(date_string)
    if len(x) == 7:
        year, month, day, hour, minute, second, utc_offset = x
        microsecond = 0
    else:
        year, month, day, hour, minute, second, microsecond, utc_offset = x

    if utc_offset:
        raise ValueError("Can't specify a time offset from UTC")

    return cftime.datetime(
        year, month, day, hour, minute, second, microsecond, calendar=calendar
    )


def st2dt(array, units_in=None):
    """The returned array is always independent.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        array: numpy array-like

        units_in: `Units`, optional

    :Returns:

        `numpy.ndarray`
            An array of `cftime.datetime` objects with the same shape
            as *array*.

    **Examples**

    """
    func = partial(st2datetime, calendar=units_in._calendar)
    return np.vectorize(func, otypes=[object])(array)


def st2rt(array, units_in, units_out):
    """The returned array is always independent.

    .. versionadded:: (cfdm) 1.11.2.0

    :Parameters:

        array: numpy array-like of ISO 8601 date-time strings

        units_in: `Units` or `None`

        units_out: `Units`

    :Returns:

        `numpy.ndarray`
            An array of floats with the same shape as *array*.

    """
    array = st2dt(array, units_in)
    return dt2rt(array, units_out)



from math import prod

import numpy as np

from ....units import Units
from ...netcdfindexer import netcdf_indexer


class FragmentArrayMixin:
    """Mixin class for a fragment of aggregated data.

    .. versionadded:: (cfdm) 1.12.0.0

    """

    def _get_array(self, index=None):
        """Returns a subspace of the dataset variable.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `__array__`, `index`

        :Parameters:

            {{index: `tuple` or `None`, optional}}

               It is important that there is a distinct value for each
               fragment dimension, which is guaranteed when the
               default of the `index` attribute is being used.

        :Returns:

            `numpy.ndarray`
                The subspace.

        """
        if index is None:
            index = self.index()

        try:
            array = super()._get_array(index)
        except ValueError:
            # A ValueError is expected to be raised when the fragment
            # variable has fewer than 'self.ndim' dimensions (we know
            # that this is the case because 'index' has 'self.ndim'
            # elements).
            axis = self._size_1_axis()  # index)
            if axis is not None:
                # There is a unique size 1 index that must correspond
                # to the missing dimension => Remove it from the
                # indices, get the fragment array with the new
                # indices; and then insert the missing size one
                # dimension.
                index = list(index)
                index.pop(axis)
                array = super()._get_array(tuple(index))
                array = np.expand_dims(array, axis)
            else:
                # There are multiple size 1 indices so we don't know
                # how many missing dimensions the fragment has, nor
                # their positions => Get the full fragment array and
                # then reshape it to the shape of the dask compute
                # chunk; and then apply the index.
                array = super()._get_array(Ellipsis)
                if array.size > prod(self.original_shape):
                    raise ValueError(
                        f"Can't get fragment data from ({self}) when "
                        "the fragment has two or more missing size 1 "
                        "dimensions, whilst also spanning two or more "
                        "Dask compute chunks."
                        "\n\n"
                        "Consider re-creating the data with exactly one "
                        "Dask compute chunk per fragment (e.g. by setting "
                        "'chunks=None' as a keyword to cf.read)."
                    )

                array = array.reshape(self.original_shape)
                array = array[index]

        array = self._conform_to_aggregated_units(array)

        # Apply any unpacking deinfed on the aggregation variable. Do
        # this after conforming the units.
        array = self._unpack_aggregated_data(array)

        return array

    def _conform_to_aggregated_units(self, array):
        """Conform the array to have the aggregated units.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            array: `numpy.ndarray` or `dict`
                The array to be conformed. If *array* is a `dict` with
                `numpy` array values then selected values are
                conformed.

        :Returns:

            `numpy.ndarray` or `dict`
                The conformed array. The returned array may or may not
                be the input array updated in-place, depending on its
                data type and the nature of its units and the
                aggregated units.

                If *array* is a `dict` then a dictionary of conformed
                arrays is returned.

        """
        units = self.Units
        if units:
            aggregated_units = self.aggregated_Units
            if not units.equivalent(aggregated_units):
                raise ValueError(
                    f"Can't convert fragment data with units {units!r} to "
                    f"have aggregated units {aggregated_units!r}"
                )

            if units != aggregated_units:
                if isinstance(array, dict):
                    # 'array' is a dictionary.
                    raise ValueError(
                        "TODOACTIVE. Placeholder notification thatn "
                        "we can't yet dealing with active "
                        "storage reductions on fragments."
                    )
                else:
                    # 'array' is a numpy array
                    array = Units.conform(
                        array, units, aggregated_units, inplace=True
                    )

        return array

    def _size_1_axis(self):  # , indices):
        """Find the position of a unique size 1 index.

        .. versionadded:: (cfdm) 1.12.0.0

        .. seealso:: `_parse_indices`, `__getitem__`

        :Parameters:

            indices: sequence of index
                The array indices to be parsed, as returned by
                `_parse_indices`.

        :Returns:

            `int` or `None`
                The position of the unique size 1 index, or `None` if
                there are zero or at least two of them.

        **Examples**

        >>> a._size_1_axis(([2, 4, 5], slice(0, 1), slice(0, 73)))
        1
        >>> a._size_1_axis(([2, 4, 5], slice(3, 4), slice(0, 73)))
        1
        >>> a._size_1_axis(([2, 4, 5], [0], slice(0, 73)))
        1
        >>> a._size_1_axis(([2, 4, 5], slice(0, 144), slice(0, 73)))
        None
        >>> a._size_1_axis(([2, 4, 5], slice(3, 7), [0, 1]))
        None
        >>> a._size_1_axis(([2, 4, 5], slice(0, 1), [0]))
        None

        """
        original_shape = self.original_shape
        if original_shape.count(1):
            return original_shape.index(1)

        return

    def _unpack_aggregated_data(self, array):
        """Unpack the canonical data, if requested.

        .. versionadded:: (cfdm) 1.12.0.0

        """
        if self.get_unpack_aggregated_data():
            array = netcdf_indexer(
                array,
                mask=False,
                unpack=True,
                attributes=self.get_aggregated_attributes(),
                copy=False,
            )[...]

        return array

    @property
    def aggregated_Units(self):
        """The units of the aggregated data.

        .. versionadded:: (cfdm) 1.12.0.0

        :Returns:

            `Units`
                The units of the aggregated data.

        """
        aggregated_attributes = self.get_aggregated_attributes(copy=False)
        calendar = aggregated_attributes.get("calendar", None)
        units = aggregated_attributes.get("units", None)
        return Units(units, calendar)

    def get_aggregated_attributes(self, copy=True):
        """The calendar of the aggregated array.

        If the calendar is `None` then the CF default calendar is
        assumed, if applicable.

        .. versionadded:: (cfdm) 1.12.0.0

        :Parameters:

            default: optional
                Return the value of the *default* parameter if the
                aggregated calendar has not been set. If set to an
                `Exception` instance then it will be raised instead.

        :Returns:

            `str` or `None`
                The calendar value.

        """
        attributes = self._get_component("aggregated_attributes")
        return attributes.copy()

    def get_unpack_aggregated_data(self):
        """Whether or not to unpack the canonical data.

        If `True` and there are aggregated variable packing
        attributes, then the array is unpacked according to those
        attributes.

        .. versionadded:: (cfdm) 1.12.0.0

        **Examples**

        >>> a.get_unpack_aggregated_data()
        True

        """
        return self._get_component("unpack_aggregated_data")



from itertools import accumulate, product

import numpy as np

from .abstract import RaggedArray
from .mixin import CompressedArrayMixin


class RaggedContiguousArray(CompressedArrayMixin, RaggedArray):
    """An underlying contiguous ragged array.

    A collection of features stored using a contiguous ragged array
    combines all features along a single dimension (the "sample
    dimension") such that each feature in the collection occupies a
    contiguous block.

    The information needed to uncompress the data is stored in a
    "count variable" that gives the size of each block.

    It is assumed that the compressed dimension is the left-most
    dimension in the compressed array.

    See CF section 9 "Discrete Sampling Geometries".

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        compressed_array=None,
        shape=None,
        size=None,
        ndim=None,
        count_variable=None,
        source=None,
        copy=True,
    ):
        """**Initialisation**

        :Parameters:

            compressed_array: array_like
                The compressed data.

            shape: `tuple`
                The shape of the uncompressed array.

            count_variable: `Count`
                The count variable required to uncompress the data,
                corresponding to a CF-netCDF count variable.

            {{init source: optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            {{init copy: `bool`, optional}}

                .. versionadded:: (cfdm) 1.10.0.0

            size: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                Number of elements in the uncompressed array.

            ndim: `int`
                Deprecated at version 1.10.0.0. Ignored if set.

                The number of uncompressed array dimensions.

        """
        super().__init__(
            compressed_array=compressed_array,
            shape=shape,
            count_variable=count_variable,
            compressed_dimensions={0: (0, 1)},
            source=source,
            copy=copy,
        )

    def subarrays(self, shapes=-1):
        """Return descriptors for every subarray.

        Theses descriptors are used during subarray decompression.

        .. versionadded:: (cfdm) 1.10.0.0

        :Parameters:

            {{subarrays chunks: ``-1`` or sequence, optional}}

        :Returns:

             4-`tuple` of iterators
                Each iterable iterates over a particular descriptor
                from each subarray.

                1. The indices of the uncompressed array that
                   correspond to each subarray.

                2. The shape of each uncompressed subarray.

                3. The indices of the compressed array that correspond
                   to each subarray.

                4. The location of each subarray on the uncompressed
                   dimensions.

        **Examples**

        An original 2-d array with shape (3, 5) comprising 3
        timeSeries features has been compressed as a contiguous ragged
        array. The features have counts of 2, 5, and 4 elements.

        >>> u_indices, u_shapes, c_indices, locations = x.subarrays()
        >>> for i in u_indices:
        ...    print(i)
        ...
        (slice(0, 1, None), slice(0, 5, None))
        (slice(1, 2, None), slice(0, 5, None))
        (slice(2, 3, None), slice(0, 5, None))
        >>> for i in u_shapes
        ...    print(i)
        ...
        (1, 5)
        (1, 5)
        (1, 5)
        >>> for i in c_indices:
        ...    print(i)
        ...
        (slice(0, 2, None),)
        (slice(2, 7, None),)
        (slice(7, 11, None),)
        >>> for i in locations:
        ...    print(i)
        ...
        (0, 0)
        (1, 0)
        (2, 0)

        """
        d1, u_dims = self.compressed_dimensions().popitem()

        shapes = self.subarray_shapes(shapes)

        # The indices of the uncompressed array that correspond to
        # each subarray, the shape of each uncompressed subarray, and
        # the location of each subarray
        locations, u_shapes, u_indices = self._uncompressed_descriptors(
            u_dims, shapes
        )

        # The indices of the compressed array that correspond to each
        # subarray
        c_indices = []
        for d, size in enumerate(self.source().shape):
            if d == d1:
                count = np.array(self.get_count()).tolist()
                c = tuple(accumulate([0] + count))
            else:
                if d < d1:
                    c = shapes[d]
                else:
                    c = shapes[d + 1]

                c = tuple(accumulate((0,) + c))

            c_indices.append([slice(i, j) for i, j in zip(c[:-1], c[1:])])

        return (
            product(*u_indices),
            product(*u_shapes),
            product(*c_indices),
            product(*locations),
        )



from . import core, mixin


class Bounds(
    mixin.NetCDFVariable,
    mixin.NetCDFDimension,
    mixin.PropertiesData,
    mixin.Files,
    core.Bounds,
):
    """A cell bounds component.

    Specifically, a cell bounds component of a coordinate or domain
    ancillary construct of the CF data model.

    An array of cell bounds spans the same domain axes as its
    coordinate array, with the addition of an extra dimension whose
    size is that of the number of vertices of each cell. This extra
    dimension does not correspond to a domain axis construct since it
    does not relate to an independent axis of the domain. Note that,
    for climatological time axes, the bounds are interpreted in a
    special way indicated by the cell method constructs.

    In the CF data model, a bounds component does not have its own
    properties because they can not logically be different to those of
    the coordinate construct itself. However, it is sometimes desired
    to store attributes on a CF-netCDF bounds variable, so it is also
    allowed to provide properties to a bounds component.

    **NetCDF interface**

    {{netCDF variable}}

    The name of the trailing netCDF dimension spanned by bounds (which
    does not correspond to a domain axis construct) may be accessed
    with the `nc_set_dimension`, `nc_get_dimension`,
    `nc_del_dimension`, and `nc_has_dimension` methods.

    {{netCDF variable group}}

    {{netCDF HDF5 chunks}}

    .. versionadded:: (cfdm) 1.7.0

    """

    def __init__(
        self,
        properties=None,
        data=None,
        source=None,
        copy=True,
        _use_data=True,
    ):
        """**Initialisation**

        :Parameters:

            {{init properties: `dict`, optional}}

                  *Parameter example:*
                     ``properties={'standard_name': 'grid_latitude'}``

            {{init data: data_like, optional}}

            {{init source: optional}}

            {{init copy: `bool`, optional}}

        """
        super().__init__(
            properties=properties,
            data=data,
            source=source,
            copy=copy,
            _use_data=_use_data,
        )

        if source is not None:
            try:
                inherited_properties = source.inherited_properties()
            except AttributeError:
                inherited_properties = {}
        else:
            inherited_properties = {}

        self._set_component(
            "inherited_properties", inherited_properties, copy=False
        )

        self._initialise_netcdf(source)
        self._initialise_original_filenames(source)

    def dump(
        self,
        display=True,
        _key=None,
        _title=None,
        _create_title=True,
        _prefix="",
        _level=0,
        _omit_properties=None,
        _axes=None,
        _axis_names=None,
    ):
        """A full description of the bounds component.

        Returns a description of all properties and provides selected
        values of all data arrays.

        .. versionadded:: (cfdm) 1.7.0

        :Parameters:

            display: `bool`, optional
                If False then return the description as a string. By
                default the description is printed.

        :Returns:

            {{returns dump}}

        """
        if _create_title and _title is None:
            _title = "Bounds: " + self.identity(default="")

        return super().dump(
            display=display,
            _key=_key,
            _omit_properties=_omit_properties,
            _prefix=_prefix,
            _level=_level,
            _title=_title,
            _create_title=_create_title,
            _axes=_axes,
            _axis_names=_axis_names,
        )

    def get_data(self, default=ValueError(), _units=True, _fill_value=True):
        """Return the data.

        Note that the data are returned in a `Data` object. Use the
        `array` attribute of the `Data` instance to return the data as an
        independent `numpy` array.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `data`, `del_data`, `has_data`, `set_data`

        :Parameters:

            default: optional
                Return the value of the *default* parameter if data have
                not been set.

                {{default Exception}}

        :Returns:

            `Data`
                The data.

        **Examples**

        >>> f = {{package}}.Field(
        ...     properties={'standard_name': 'surface_altitude'})
        >>> d = {{package}}.Data(range(10))
        >>> f.set_data(d)
        >>> f.has_data()
        True
        >>> f.get_data()
        <{{repr}}Data(10): [0, ..., 9]>

        >>> f.del_data()
        <{{repr}}Data(10): [0, ..., 9]>
        >>> f.has_data()
        False
        >>> print(f.get_data(None))
        None
        >>> print(f.del_data(None))
        None

        """
        data = super().get_data(
            default=None, _units=_units, _fill_value=_fill_value
        )
        if data is None:
            if default is None:
                return

            return self._default(
                default, f"{self.__class__.__name__} has no data"
            )

        if _units or _fill_value:
            inherited_properties = self._get_component(
                "inherited_properties", {}
            )

            if _units:
                if not data.has_units():
                    units = inherited_properties.get("units")
                    if units is not None:
                        data.set_units(units)

                if not data.has_calendar():
                    calendar = inherited_properties.get("calendar")
                    if calendar is not None:
                        data.set_calendar(calendar)

            if _fill_value and not data.has_fill_value():
                fv = inherited_properties.get("fill_value")  # TODO
                if fv is not None:
                    data.set_fill_value(fv)

        return data

    def inherited_properties(self):
        """Return the properties inherited from a coordinate construct.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `properties`

        :Returns:

            `dict`
                The inherited properties.

        **Examples**

        >>> f = {{package}}.example_field(6)
        >>> d = f.constructs('longitude').value()
        >>> b = d.bounds
        >>> b
        <{{repr}}Bounds: longitude(2, 3, 4) degrees_east>

        >>> b.inherited_properties()
        {'units': 'degrees_east', 'standard_name': 'longitude'}

        """
        return self._get_component("inherited_properties", {}).copy()

    def identity(self, default=""):
        """Return the canonical identity.

        By default the identity is the first found of the following:

        1. The ``standard_name`` property.
        2. The ``cf_role`` property, preceded by ``'cf_role='``.
        3. The ``long_name`` property, preceded by ``'long_name='``.
        4. The netCDF variable name, preceded by ``'ncvar%'``.
        5. The value of the *default* parameter.

        Properties include any inherited properties.

        .. versionadded:: (cfdm) 1.7.0

        .. seealso:: `identities`

        :Parameters:

            default: optional
                If no identity can be found then return the value of the
                default parameter.

        :Returns:

                The identity.

        **Examples**

        >>> f = {{package}}.example_field(6)
        >>> d = f.constructs('longitude').value()
        >>> b = d.bounds
        >>> b
        <{{repr}}Bounds: longitude(2, 3, 4) degrees_east>
        >>> b.identity()
        'longitude'

        """
        inherited_properties = self._get_component(
            "inherited_properties", None
        )
        if inherited_properties:
            bounds = self.copy()
            properties = bounds.properties()
            bounds.set_properties(inherited_properties)
            bounds.set_properties(properties)
            self = bounds

        return super().identity(default=default)
